from .. import main
from types import ModuleType
from typing import Union


def find_object(context: Union[ModuleType, dict], attribute: str, prop_name: str = None):
    """
    Helper function to resolve an object matching the name attribute
    inside the context. If it's not found, throws ValueError suggesting
    the user to override the `nnfabrik` class property or to a specific
    class property for the table.

    Args:
        context (Union[ModuleType, dict]): A context object in which the name attribute would be checked.
            Can either be a module object or a dictionary.
        attribute (str): Name of object being sought.
        prop_name (str, optional): The property name under which this object is being sought. Defaults to None,
            in which case the name is infered to be lower(attribute) + '_table'. E.g. `model_table` for attribute
            'Model'.

    Raises:
        ValueError: if an object with name `attribute` is not found inside the context.

    Returns:
        Any: the object with name `attribute` found inside the context.
    """
    # if context of string "core" given, then use the core main module as the context
    if context == "core":
        context = main

    if prop_name is None:
        prop_name = attribute.lower() + "_table"

    if context is None:
        raise ValueError("Please specify either `nnfabrik` or `{}` property for the class".format(prop_name))

    if isinstance(context, ModuleType):
        context = context.__dict__

    return context[attribute]



import datajoint as dj
import tempfile
import torch
import os
from nnfabrik.main import Model, Dataset, Trainer, Seed, Fabrikant
from nnfabrik.utility.dj_helpers import gitlog, make_hash
from .trained_model import TrainedModelBase


class TransferredTrainedModelBase(TrainedModelBase):
    """
    A modified version of TrainedModel table which enables step-wise table population given the
    step specification. Refer to the corresponding example notebook for a demo.
    """

    table_comment = "Transferred trained models"

    @property
    def definition(self):
        definition = """
        # {table_comment}
        transfer_step:                     int          # transfer step
        -> self.model_table
        -> self.dataset_table
        -> self.trainer_table
        -> self.seed_table
        prev_model_fn:                     varchar(64)
        prev_model_hash:                   varchar(64)
        prev_dataset_fn:                   varchar(64)
        prev_dataset_hash:                 varchar(64)
        prev_trainer_fn:                   varchar(64)
        prev_trainer_hash:                 varchar(64)
        ---
        comment='':                        varchar(768) # short description 
        score:                             float        # loss
        output:                            longblob     # trainer object's output
        ->[nullable] self.user_table
        trainedmodel_ts=CURRENT_TIMESTAMP: timestamp    # UTZ timestamp at time of insertion
        current_model_fn:                  varchar(64)
        current_model_hash:                varchar(64)
        current_dataset_fn:                varchar(64)
        current_dataset_hash:              varchar(64)
        current_trainer_fn:                varchar(64)
        current_trainer_hash:              varchar(64)
        """.format(
            table_comment=self.table_comment
        )
        return definition

    class ModelStorage(TrainedModelBase.ModelStorage):
        pass

    def _transfer_recipe(self, transfer_step):
        """
        Combines multiple transfer recipes and their resitrictions as specified by post_restr attribute.
        The combination is transfer-step-specific, meaning only the recipes relevant for a specific transfer step would be combined.

        Combining recipes are pretty easy and the user does not need to interact with this method directly. Below is an example:
        Let us assume you have two recipe tables: TrainerRecipe and ModelRecipe, the you can attach all this recipes to your
        TransferTrainedModel table as follow:

        ``` Python
            TransferTrainedModel.transfer_recipe = [TrainerRecipe, ModelRecipe]
        ```

        The rest (combining the recipes and their restrictions) is taken care of by this method.

        Args:
            transfer_step (int): table population trasnfer step.

        Returns:
            string or datajoint AndList: A single or combined restriction of one or multiple recipes, respectively.
        """

        if isinstance(self.transfer_recipe, list):

            # get the recipes that have a entry for a specific transfer step
            # transfer_recipe = [tr & f"transfer_step = {transfer_step}" for tr in self.transfer_recipe if tr & f"transfer_step = {transfer_step}"]
            transfer_recipe = []
            # loop over the transfer recipes
            for tr in self.transfer_recipe:
                # check if an entry exists for a specific transfer step in the recipe
                if tr & f"transfer_step = {transfer_step}":
                    # if it exists add that entry to the list of recipes (relevant for a specific transfer step)
                    transfer_recipe.append(tr & f"transfer_step = {transfer_step}")

            # join all the recipes (and their post_restr)
            joined = transfer_recipe[0]

            if len(transfer_recipe) > 1:
                for t in transfer_recipe[1:]:
                    joined *= t

                joined.post_restr = dj.AndList([recipe.post_restr for recipe in self.transfer_recipe])

            return joined

        else:
            return self.transfer_recipe

    @property
    def key_source(self):

        if hasattr(self, "transfer_recipe"):

            # project (rename) attributes of the existing transfereedmodel table to the same name but with prefix "prev"
            prev_transferredmodel = self.proj(
                prev_model_fn="current_model_fn",
                prev_model_hash="current_model_hash",
                prev_dataset_fn="current_dataset_fn",
                prev_dataset_hash="current_dataset_hash",
                prev_trainer_fn="current_trainer_fn",
                prev_trainer_hash="current_trainer_hash",
                prev_step="transfer_step",
                transfer_step="transfer_step + 1",
            ) * dj.U(
                "transfer_step",  # make these attributes primary keys
                "prev_model_fn",
                "prev_model_hash",
                "prev_dataset_fn",
                "prev_dataset_hash",
                "prev_trainer_fn",
                "prev_trainer_hash",
            )

            # get the current transfer step
            transfer_step = prev_transferredmodel.fetch("transfer_step").max() if prev_transferredmodel else 0

            if transfer_step:

                # get the necessay attributes to filter the prev_transferredmodel with the transfer recipe
                prev_transferredmodel = (
                    dj.U(
                        "transfer_step",
                        "prev_model_fn",
                        "prev_model_hash",
                        "prev_dataset_fn",
                        "prev_dataset_hash",
                        "prev_trainer_fn",
                        "prev_trainer_hash",
                    )
                    & prev_transferredmodel
                )

                # get the entries that match the one in TransferRecipe (for specification of previous)
                transfer_from = prev_transferredmodel * self._transfer_recipe(transfer_step)

                transfers = (
                    dj.U(
                        "transfer_step",
                        "model_fn",
                        "model_hash",
                        "dataset_fn",
                        "dataset_hash",
                        "trainer_fn",
                        "trainer_hash",
                        "seed",
                        "prev_model_fn",
                        "prev_model_hash",
                        "prev_dataset_fn",
                        "prev_dataset_hash",
                        "prev_trainer_fn",
                        "prev_trainer_hash",
                    )
                    & Model * Dataset * Trainer * Seed * transfer_from
                    & self._transfer_recipe(transfer_step).post_restr
                )

                return transfers.proj()

            else:

                # set of models to be trained from scratch (I arbitrary chose to only train if trainer_id + model_id < 3)
                step_0 = Model * Dataset * Trainer * Seed

                # add transfer_step and prev_hash as prim keys
                base = dj.U(
                    "transfer_step",
                    "prev_model_fn",
                    "prev_model_hash",
                    "prev_dataset_fn",
                    "prev_dataset_hash",
                    "prev_trainer_fn",
                    "prev_trainer_hash",
                ) * step_0.proj(
                    transfer_step="0",
                    prev_model_fn='""',
                    prev_model_hash='""',
                    prev_dataset_fn='""',
                    prev_dataset_hash='""',
                    prev_trainer_fn='""',
                    prev_trainer_hash='""',
                )
                return base.proj()

        else:
            # normal entries as a combinatio of Dataset, Model, Trainer, and Seed tables
            step_0 = Model * Dataset * Trainer * Seed

            # add transfer_step and prev_hash as prim keys
            base = dj.U(
                "transfer_step",
                "prev_model_fn",
                "prev_model_hash",
                "prev_dataset_fn",
                "prev_dataset_hash",
                "prev_trainer_fn",
                "prev_trainer_hash",
            ) * step_0.proj(
                transfer_step="0",
                prev_model_fn='""',
                prev_model_hash='""',
                prev_dataset_fn='""',
                prev_dataset_hash='""',
                prev_trainer_fn='""',
                prev_trainer_hash='""',
            )
            return base.proj()

    def make(self, key):
        """
        Given key specifying configuration for dataloaders, model and trainer,
        trains the model and saves the trained model.
        """

        # lookup the fabrikant corresponding to the current DJ user
        fabrikant_name = Fabrikant.get_current_user()
        seed = (Seed & key).fetch1("seed")

        # load everything
        dataloaders, model, trainer = self.load_model(key, include_trainer=True, include_state_dict=False, seed=seed)

        # define callback with pinging
        def call_back(**kwargs):
            self.connection.ping()
            self.call_back(**kwargs)

        # model training
        score, output, model_state = trainer(model=model, dataloaders=dataloaders, seed=seed, uid=key, cb=call_back)

        with tempfile.TemporaryDirectory() as temp_dir:
            filename = make_hash(key) + ".pth.tar"
            filepath = os.path.join(temp_dir, filename)
            torch.save(model_state, filepath)

            key["score"] = score
            key["output"] = output
            key["fabrikant_name"] = fabrikant_name
            comments = []
            comments.append((self.trainer_table & key).fetch1("trainer_comment"))
            comments.append((self.model_table & key).fetch1("model_comment"))
            comments.append((self.dataset_table & key).fetch1("dataset_comment"))
            key["comment"] = self.comment_delimitter.join(comments)

            key["current_model_fn"], key["current_model_hash"] = (Model & key).fetch1("model_fn", "model_hash")
            key["current_dataset_fn"], key["current_dataset_hash"] = (Dataset & key).fetch1(
                "dataset_fn", "dataset_hash"
            )
            key["current_trainer_fn"], key["current_trainer_hash"] = (Trainer & key).fetch1(
                "trainer_fn", "trainer_hash"
            )

            self.insert1(key)

            key["model_state"] = filepath

            self.ModelStorage.insert1(key, ignore_extra_fields=True)



from typing import Dict, Tuple, Callable, Optional, List, Any

from tqdm import tqdm
import torch
import torch.nn as nn

from nnfabrik.examples.mnist.trainer import MNISTTrainer


class ChkptTrainer(MNISTTrainer):
    def __init__(
        self,
        model: nn.Module,
        dataloaders: Dict,
        seed: int,
        uid: Tuple,
        call_back: Callable,
        epochs: int = 5,
        chkpt_options: Optional[Dict] = None,
    ) -> None:
        super(ChkptTrainer, self).__init__(model, dataloaders, seed, epochs)
        self.trained_model_cb = call_back
        self.uid = uid
        self.accs = []
        self.chkpt_options = chkpt_options if chkpt_options is not None else {}

    def save(self, epoch: int, score: float) -> None:
        state = {
            "action": "save",
            "score": score,
            "maximize_score": True,
            "tracker": self.accs,
            "optimizer": self.optimizer,
            **self.chkpt_options,
        }
        self.trained_model_cb(
            uid=self.uid,
            epoch=epoch,
            model=self.model,
            state=state,
        )  # save model

    def restore(self) -> int:
        loaded_state = {
            "action": "last",
            "maximize_score": True,
            "tracker": self.accs,
            "optimizer": self.optimizer.state_dict(),
        }
        self.trained_model_cb(
            uid=self.uid, epoch=-1, model=self.model, state=loaded_state
        )  # load the last epoch if existing
        epoch = loaded_state.get("epoch", -1) + 1
        return epoch

    def train(self) -> Tuple[float, Tuple[List[float], int], Dict]:
        if hasattr(tqdm, "_instances"):
            tqdm._instances.clear()  # To have tqdm output without line-breaks between steps
        torch.manual_seed(self.seed)
        start_epoch = self.restore()
        for epoch in range(start_epoch, self.epochs):
            print(f"Epoch {epoch}")
            predicted_correct = 0
            total = 0
            for x, y in tqdm(self.trainloader):
                p, t = self.train_loop(x, y)
                predicted_correct += p
                total += t
            self.accs.append(100.0 * predicted_correct / total)
            self.save(epoch, self.accs[-1])

        return self.accs[-1], (self.accs, self.epochs), self.model.state_dict()


def chkpt_trainer_fn(
    model: torch.nn.Module,
    dataloaders: Dict,
    seed: int,
    uid: Dict,
    cb: Callable,
    **config,
) -> Tuple[float, Any, Dict]:
    """ "
    Args:
        model: initialized model to train
        data_loaders: containing "train", "validation" and "test" data loaders
        seed: random seed
        uid: database keys that uniquely identify this trainer call
        cb: callback function to ping the database and potentially save the checkpoint
    Returns:
        score: performance score of the model
        output: user specified validation object based on the 'stop function'
        model_state: the full state_dict() of the trained model
    """
    trainer = ChkptTrainer(model, dataloaders, seed, uid=uid, call_back=cb, epochs=config.get("epochs", 2))
    out = trainer.train()

    return out



from typing import Dict, Tuple, Callable, List, Any

from tqdm import tqdm
import torch
from torch import nn
from torch import optim


class MNISTTrainer:
    def __init__(
        self,
        model: nn.Module,
        dataloaders: Dict,
        seed: int,
        epochs: int = 5,
    ) -> None:

        self.model = model
        self.trainloader = dataloaders["train"]
        self.seed = seed
        self.epochs = epochs
        self.loss_fn = nn.NLLLoss()
        self.optimizer = optim.Adam(self.model.parameters())

    def train_loop(self, x: torch.Tensor, y: torch.Tensor) -> Tuple[float, int]:
        # forward:
        self.optimizer.zero_grad()
        x_flat = x.flatten(1, -1)  # treat the images as flat vectors
        logits = self.model(x_flat)
        loss = self.loss_fn(logits, y)
        # backward:
        loss.backward()
        self.optimizer.step()
        # keep track of accuracy:
        _, predicted = logits.max(1)
        predicted_correct = predicted.eq(y).sum().item()
        total = y.shape[0]
        return predicted_correct, total

    def train(self) -> Tuple[float, Tuple[List[float], int], Dict]:
        if hasattr(tqdm, "_instances"):
            tqdm._instances.clear()  # To have tqdm output without line-breaks between steps
        torch.manual_seed(self.seed)
        accs = []
        for epoch in range(self.epochs):
            predicted_correct = 0
            total = 0
            for x, y in tqdm(self.trainloader):
                p, t = self.train_loop(x, y)
                predicted_correct += p
                total += t

            accs.append(100.0 * predicted_correct / total)

        return accs[-1], (accs, self.epochs), self.model.state_dict()


def mnist_trainer_fn(
    model: torch.nn.Module, dataloaders: Dict, seed: int, uid: Dict, cb: Callable, **config
) -> Tuple[float, Any, Dict]:
    """
    Args:
        model: initialized model to train
        data_loaders: containing "train", "validation" and "test" data loaders
        seed: random seed
        uid: database keys that uniquely identify this trainer call
        cb: callback function to ping the database and potentially save the checkpoint
    Returns:
        score: performance score of the model
        output: user specified validation object based on the 'stop function'
        model_state: the full state_dict() of the trained model
    """
    trainer = MNISTTrainer(model, dataloaders, seed, epochs=config.get("epochs", 2))
    out = trainer.train()

    return out



# importing the tables here is a trick to get IntelliSense to work
from nnfabrik.main import Fabrikant, Trainer, Dataset, Model, Seed, my_nnfabrik

# define nnfabrik tables here
my_nnfabrik(
    "nnfabrik_example",
    context=locals(),
)



import warnings
import types
from typing import Union, Optional, MutableMapping, Tuple

import datajoint as dj

from .builder import (
    resolve_model,
    resolve_data,
    resolve_trainer,
    get_data,
    get_model,
    get_trainer,
)
from .utility.dj_helpers import make_hash, CustomSchema, Schema
from .utility.nnf_helper import cleanup_numpy_scalar


schema = CustomSchema(dj.config.get("nnfabrik.schema_name", "nnfabrik_core"))


@schema
class Fabrikant(dj.Manual):
    definition = """
    fabrikant_name: varchar(32)       # Name of the contributor that added this entry
    ---
    full_name="": varchar(128) # full name of the person
    email: varchar(64)         # e-mail address
    affiliation: varchar(32)   # conributor's affiliation (e.g. Sinz Lab)
    dj_username: varchar(64)   # DataJoint username
    """

    def add_entry(
        self,
        name,
        email,
        affiliation,
        full_name="",
        dj_username=None,
        skip_duplicates=False,
        return_pk_only=True,
    ):
        """
        Add a new entry into Fabrikant table. If `dj_username` is omitted, then the current
        database connection user is used.

        Args:
            name (str): A short name to identify yourself.
            email (str): Email address.
            affiliation (str): Lab or institutional affiliation.
            full_name (str, optional): Full name. Defaults to "".
            dj_username (str, optional): DataJoint username. Defaults to None, in which case
                the username of the current connection is used.
            skip_duplicates (bool, optional): If True, no error is thrown when a duplicate entry (i.e. entry with same model_fn and model_config) is found. Defaults to False.
            return_pk_only (bool, optional): If True, only the primary key attribute for the new entry or corresponding existing entry is returned. Otherwise, the entire
                entry is returned. Defaults to True.

        Returns:
            dict: the entry in the table corresponding to the new (or possibly existing, if skip_duplicates=True) entry.
        """
        if dj_username is None:
            dj_username = self.connection.get_user().split("@")[0]

        key = dict(
            fabrikant_name=name,
            full_name=full_name,
            email=email,
            affiliation=affiliation,
            dj_username=dj_username,
        )

        # overlap in DJ username is not allowed either
        existing = self.proj() & key or (self & dict(dj_username=dj_username)).proj()
        if existing:
            key = (self & (existing)).fetch1()
            if skip_duplicates:
                warnings.warn("Corresponding entry found. Skipping...")
            else:
                raise ValueError("Corresponding entry already exists: {}".format(key))
        else:
            self.insert1(key, ignore_extra_fields=True)

        if return_pk_only:
            key = {k: key[k] for k in self.heading.primary_key}

        return key

    @classmethod
    def get_current_user(cls):
        """
        Lookup the fabrikant_name in Fabrikant corresponding to the currently logged in DataJoint user
        Returns: fabrikant_name if match found, else None
        """
        username = cls.connection.get_user().split("@")[0]
        entry = Fabrikant & dict(dj_username=username)
        if entry:
            return entry.fetch1("fabrikant_name")


@schema
class Model(dj.Manual):
    definition = """
    model_fn:                   varchar(64)   # name of the model function
    model_hash:                 varchar(64)   # hash of the model configuration
    ---
    model_config:               longblob      # model configuration to be passed into the function
    -> Fabrikant.proj(model_fabrikant='fabrikant_name')
    model_comment='' :          varchar(256)   # short description
    model_ts=CURRENT_TIMESTAMP: timestamp     # UTZ timestamp at time of insertion
    """

    @property
    def fn_config(self):
        model_fn, model_config = self.fetch1("model_fn", "model_config")
        model_config = cleanup_numpy_scalar(model_config)
        return model_fn, model_config

    @staticmethod
    def resolve_fn(fn_name):
        return resolve_model(fn_name)

    def add_entry(
        self,
        model_fn,
        model_config,
        model_comment="",
        model_fabrikant=None,
        skip_duplicates=False,
        return_pk_only=True,
    ):
        """
        Add a new entry to the model.

        Args:
            model_fn (str, Callable): name of a callable object. If name contains multiple parts separated by `.`, this is assumed to be found in a another module and
                dynamic name resolution will be attempted. Other wise, the name will be checked inside `models` subpackage.
            model_config (dict): Python dictionary containing keyword arguments for the model_fn
            model_comment - Optional comment for the entry.
            model_fabrikant (str): The fabrikant name. Must match an existing entry in Fabrikant table. If ignored, will attempt to resolve Fabrikant based on the database user name for the existing connection.
            skip_duplicates (bool, optional): If True, no error is thrown when a duplicate entry (i.e. entry with same model_fn and model_config) is found. Defaults to False.
            return_pk_only (bool, optional): If True, only the primary key attribute for the new entry or corresponding existing entry is returned. Otherwise, the entire
                entry is returned. Defaults to True.

        Returns:
            dict: the entry in the table corresponding to the new (or possibly existing, if skip_duplicates=True) entry.
        """
        if not isinstance(model_fn, str):
            # infer the full path to the callable
            model_fn = model_fn.__module__ + "." + model_fn.__name__

        try:
            resolve_model(model_fn)
        except (NameError, TypeError) as e:
            warnings.warn(str(e) + "\nTable entry rejected")
            return

        if model_fabrikant is None:
            model_fabrikant = Fabrikant.get_current_user()

        model_hash = make_hash(model_config)
        key = dict(
            model_fn=model_fn,
            model_hash=model_hash,
            model_config=model_config,
            model_fabrikant=model_fabrikant,
            model_comment=model_comment,
        )

        existing = self.proj() & key
        if existing:
            if skip_duplicates:
                warnings.warn("Corresponding entry found. Skipping...")
                key = (self & (existing)).fetch1()
            else:
                raise ValueError("Corresponding entry already exists")
        else:
            self.insert1(key)

        if return_pk_only:
            key = {k: key[k] for k in self.heading.primary_key}

        return key

    def build_model(self, dataloaders=None, seed=None, key=None, data_info=None):
        """
        Builds a Pytorch module by calling the model_fn with the corresponding model_config. The table has to be
        restricted to one entry in order for this method to return a model.
        Either the dataloaders or data_info have to be specified to determine the size of the input and thus the
        appropriate model settings.

        Args:
            dataloaders (dict) -  a dictionary of dataloaders. The model will infer its shape from these dataloaders
            seed (int) -  random seed
            key (dict) - datajoint key
            data_info (dict) - contains all necessary information about the input in order to build the model.

        Returns:
            A PyTorch module.
        """
        if dataloaders is None and data_info is None:
            raise ValueError(
                "dataloaders and data_info can not both be None. To build the model, either dataloaders or"
                "data_info have to be passed."
            )

        print("Building model...")
        if key is None:
            key = {}
        model_fn, model_config = (self & key).fn_config

        return get_model(
            model_fn,
            model_config,
            dataloaders=dataloaders,
            seed=seed,
            data_info=data_info,
        )


@schema
class Dataset(dj.Manual):
    definition = """
    dataset_fn:                     varchar(64)    # name of the dataset loader function
    dataset_hash:                   varchar(64)    # hash of the configuration object
    ---
    dataset_config:                 longblob       # dataset configuration object
    -> Fabrikant.proj(dataset_fabrikant='fabrikant_name')
    dataset_comment='' :            varchar(256)    # short description
    dataset_ts=CURRENT_TIMESTAMP:   timestamp      # UTZ timestamp at time of insertion
    """

    @property
    def fn_config(self):
        dataset_fn, dataset_config = self.fetch1("dataset_fn", "dataset_config")
        dataset_config = cleanup_numpy_scalar(dataset_config)
        return dataset_fn, dataset_config

    @staticmethod
    def resolve_fn(fn_name):
        return resolve_data(fn_name)

    def add_entry(
        self,
        dataset_fn,
        dataset_config,
        dataset_comment="",
        dataset_fabrikant=None,
        skip_duplicates=False,
        return_pk_only=True,
    ):
        """
        Add a new entry to the dataset.

        Args:
            dataset_fn (string, Callable): name of a callable object. If name contains multiple parts separated by `.`, this is assumed to be found in a another module and
                dynamic name resolution will be attempted. Other wise, the name will be checked inside `models` subpackage.
            dataset_config (dict): Python dictionary containing keyword arguments for the dataset_fn
            dataset_comment (str, optional):  Comment for the entry. Defaults to "" (emptry string)
            dataset_fabrikant (string): The fabrikant name. Must match an existing entry in Fabrikant table. If ignored, will attempt to resolve Fabrikant based
                on the database user name for the existing connection.
            skip_duplicates (bool, optional): If True, no error is thrown when a duplicate entry (i.e. entry with same model_fn and model_config) is found. Defaults to False.
            return_pk_only (bool, optional): If True, only the primary key attribute for the new entry or corresponding existing entry is returned. Otherwise, the entire
                entry is returned. Defaults to True.

        Returns:
            dict: the entry in the table corresponding to the new (or possibly existing, if skip_duplicates=True) entry.
        """
        if not isinstance(dataset_fn, str):
            # infer the full path to the callable
            dataset_fn = dataset_fn.__module__ + "." + dataset_fn.__name__

        try:
            resolve_data(dataset_fn)
        except (NameError, TypeError) as e:
            warnings.warn(str(e) + "\nTable entry rejected")
            return

        if dataset_fabrikant is None:
            dataset_fabrikant = Fabrikant.get_current_user()

        dataset_hash = make_hash(dataset_config)
        key = dict(
            dataset_fn=dataset_fn,
            dataset_hash=dataset_hash,
            dataset_config=dataset_config,
            dataset_fabrikant=dataset_fabrikant,
            dataset_comment=dataset_comment,
        )

        existing = self.proj() & key
        if existing:
            if skip_duplicates:
                warnings.warn("Corresponding entry found. Skipping...")
                key = (self & (existing)).fetch1()
            else:
                raise ValueError("Corresponding entry already exists")
        else:
            self.insert1(key)

        if return_pk_only:
            key = {k: key[k] for k in self.heading.primary_key}

        return key

    def get_dataloader(self, seed=None, key=None):
        """
        Returns a dataloader for a given dataset loader function and its corresponding configurations
        dataloader: is expected to be a dict in the form of
                            {
                            'train': torch.utils.data.DataLoader,
                            'val': torch.utils.data.DataLoader,
                            'test: torch.utils.data.DataLoader,
                             }
                             or a similar iterable object
                each loader should have as first argument the input such that
                    next(iter(train_loader)): [input, responses, ...]
                the input should have the following form:
                    [batch_size, channels, px_x, px_y, ...]
        """
        # TODO: update the docstring

        if key is None:
            key = {}

        dataset_fn, dataset_config = (self & key).fn_config

        if seed is not None:
            dataset_config["seed"] = seed  # override the seed if passed in

        return get_data(dataset_fn, dataset_config)


@schema
class Trainer(dj.Manual):
    definition = """
    trainer_fn:                     varchar(64)     # name of the Trainer loader function
    trainer_hash:                   varchar(64)     # hash of the configuration object
    ---
    trainer_config:                 longblob        # training configuration object
    -> Fabrikant.proj(trainer_fabrikant='fabrikant_name')
    trainer_comment='' :            varchar(256)     # short description
    trainer_ts=CURRENT_TIMESTAMP:   timestamp       # UTZ timestamp at time of insertion
    """

    @property
    def fn_config(self):
        trainer_fn, trainer_config = self.fetch1("trainer_fn", "trainer_config")
        trainer_config = cleanup_numpy_scalar(trainer_config)
        return trainer_fn, trainer_config

    @staticmethod
    def resolve_fn(fn_name):
        return resolve_trainer(fn_name)

    def add_entry(
        self,
        trainer_fn,
        trainer_config,
        trainer_comment="",
        trainer_fabrikant=None,
        skip_duplicates=False,
        return_pk_only=True,
    ):
        """
        Add a new entry to the trainer.

        Args:
            trainer_fn (str): name of a callable object. If name contains multiple parts separated by `.`, this is assumed to be found in a another module and
                dynamic name resolution will be attempted. Other wise, the name will be checked inside `models` subpackage.
            trainer_config (dict): Python dictionary containing keyword arguments for the trainer_fn.
            trainer_comment (str, optional): Optional comment for the entry. Defaults to "" (empty string).
            trainer_fabrikant (str): The fabrikant name. Must match an existing entry in Fabrikant table. If ignored, will attempt to resolve Fabrikant based
                on the database user name for the existing connection.
            skip_duplicates (bool, optional): If True, no error is thrown when a duplicate entry (i.e. entry with same model_fn and model_config) is found. Defaults to False.
            return_pk_only (bool, optional): If True, only the primary key attribute for the new entry or corresponding existing entry is returned. Otherwise, the entire
                entry is returned. Defaults to True.

        Returns:
            dict: the entry in the table corresponding to the new (or possibly existing, if skip_duplicates=True) entry.
        """
        if not isinstance(trainer_fn, str):
            # infer the full path to the callable
            trainer_fn = trainer_fn.__module__ + "." + trainer_fn.__name__

        try:
            resolve_trainer(trainer_fn)
        except (NameError, TypeError) as e:
            warnings.warn(str(e) + "\nTable entry rejected")
            return

        if trainer_fabrikant is None:
            trainer_fabrikant = Fabrikant.get_current_user()

        trainer_hash = make_hash(trainer_config)
        key = dict(
            trainer_fn=trainer_fn,
            trainer_hash=trainer_hash,
            trainer_config=trainer_config,
            trainer_fabrikant=trainer_fabrikant,
            trainer_comment=trainer_comment,
        )

        existing = self.proj() & key
        if existing:
            if skip_duplicates:
                warnings.warn("Corresponding entry found. Skipping...")
                key = (self & (existing)).fetch1()
            else:
                raise ValueError("Corresponding entry already exists")
        else:
            self.insert1(key)

        if return_pk_only:
            key = {k: key[k] for k in self.heading.primary_key}

        return key

    def get_trainer(self, key=None, build_partial=True):
        """
        Returns the trainer function and its corresponding configurations. If build_partial=True (default), then it constructs
        a partial function with configuration object already passed in, thus returning only a single function.
        """
        if key is None:
            key = {}
        trainer_fn, trainer_config = (self & key).fn_config

        if build_partial:
            # build the configuration into the function
            return get_trainer(trainer_fn, trainer_config)
        else:
            # return them separately
            return get_trainer(trainer_fn), trainer_config


@schema
class Seed(dj.Manual):
    definition = """
    seed:   int     # Random seed that is passed to the model- and dataset-builder
    """


def my_nnfabrik(
    schema: Union[str, Schema],
    additional_tables: Tuple = (),
    use_common_fabrikant: bool = True,
    use_common_seed: bool = False,
    module_name: Optional[str] = None,
    context: Optional[MutableMapping] = None,
    spawn_existing_tables: bool = False,
) -> Optional[types.ModuleType]:
    """
    Create a custom nnfabrik module under specified DataJoint schema,
    instantitaing Model, Dataset, and Trainer tables. If `use_common_fabrikant`
    is set to True, the new tables will depend on the common Fabrikant table.
    Otherwise, a separate copy of Fabrikant table will also be prepared.

    Examples:
        Use of this function should replace any existing use of `nnfabrik` tables done via modifying the
        `nnfabrik.schema_name` property in `dj.config`.

        As an example, if you previously had a code like this:
        >>> dj.config['nfabrik.schema_name'] = 'my_schema'
        >>> from nnfabrik import main # importing nnfabrik tables

        do this instead:
        >>> from nnfabrik.main import my_nnfabrik
        >>> main = my_nnfabrik('my_schema')    # this has the same effect as defining nnfabrik tables in schema `my_schema`

        Also, you can achieve the equivalent of:
        >>> dj.config['nfabrik.schema_name'] = 'my_schema'
        >>> from nnfabrik.main import *

        by doing
        >>> from nnfabrik.main import my_nnfabrik
        >>> my_nnfabrik('my_schema', context=locals())

    Args:
        schema (str or dj.Schema): Name of schema or dj.Schema object
        use_common_fabrikant (bool, optional): If True, new tables will depend on the
           common Fabrikant table. If False, new copy of Fabrikant will be created and used.
           Defaults to True.
        use_common_seed (bool, optional): If True, new tables will depend on the
           common Seed table. If False, new copy of Seed will be created and used.
           Defaults to False.
        module_name (str, optional): Name property of the returned Python module object.
            Defaults to None, in which case the name of the schema will be used.
        context (dict, optional): If non None value is provided, then a module is not created and
            instead the tables are defined inside the context.
        spawn_existing_tables (bool, optional): If True, perform `spawn_missing_tables` operation
            onto the newly created table. Defaults to False.

    Raises:
        ValueError: If `use_common_fabrikant` is True but the target `schema` already contains its own
            copy of `Fabrikant` table, or if `use_common_seed` is True but the target `schema` already
            contains its own copy of `Seed` table.

    Returns:
        Python Module object or None: If `context` was None, a new Python module containing
            nnfabrik tables defined under the schema. The module's schema property points
            to the schema object as well. Otherwise, nothing is returned.
    """
    if isinstance(schema, str):
        schema = CustomSchema(schema)

    tables = [Seed, Fabrikant, Model, Dataset, Trainer] + list(additional_tables)

    module = None
    if context is None:
        module_name = schema.database if module_name is None else module_name
        module = types.ModuleType(module_name)
        context = module.__dict__

    context["schema"] = schema

    # spawn all existing tables into the module
    # TODO: replace with a cheaper check operation
    temp_context = context if spawn_existing_tables else {}
    schema.spawn_missing_classes(temp_context)

    if use_common_fabrikant:
        if "Fabrikant" in temp_context:
            raise ValueError(
                "The schema already contains a Fabrikant table despite setting use_common_fabrikant=True. "
                "Either rerun with use_common_fabrikant=False or remove the Fabrikant table in the schema"
            )
        context["Fabrikant"] = Fabrikant
        # skip creating Fabrikant table
        tables.remove(Fabrikant)

    if use_common_seed:
        if "Seed" in temp_context:
            raise ValueError(
                "The schema already contains a Seed table despite setting use_common_seed=True. "
                "Either rerun with use_common_seed=False or remove the Seed table in the schema"
            )
        context["Seed"] = Seed
        # skip creating Seed table
        tables.remove(Seed)

    for table in tables:
        new_table = type(table.__name__, (table,), dict(__doc__=table.__doc__))
        context[table.__name__] = schema(new_table, context=context)

    # this returns None if context was set
    return module



# helper functions for use with DataJoint tables

import warnings
from datetime import datetime
import hashlib
import datajoint as dj
from git import Repo, cmd
import numpy as np
import inspect
from datetime import date, datetime
from datajoint.utils import to_camel_case
from collections import OrderedDict, Iterable, Mapping

# try/except is necessary to support all versions of dj
try:
    # for versions < 0.12.6
    from datajoint.schema import Schema
except:
    # for versions >= 0.12.6
    from datajoint.schemas import Schema


def clone_conn(conn):
    """
    Clones a given connection object.
    Args:
        conn (dj.Connection): connection object to clone
    Returns:
        new dj.Connection object with properties identical to conn
    """
    conn_info = conn.conn_info
    return dj.Connection(
        host=conn_info["host"],
        user=conn_info["user"],
        password=conn_info["passwd"],
        init_fun=conn.init_fun,
        use_tls=conn_info["ssl"],
    )


def cleanup_numpy_scalar(data):
    """
    Recursively cleanups up a (potentially nested data structure of)
    objects, replacing any scalar numpy instance with the corresponding
    Python native datatype.
    """
    if isinstance(data, np.generic):
        if data.shape == ():
            data = data.item()
    elif isinstance(data, dict):
        for k, v in data.items():
            data[k] = cleanup_numpy_scalar(v)
    elif isinstance(data, (list, tuple)):
        data = [cleanup_numpy_scalar(e) for e in data]
    return data


def make_hash(obj):
    """
    Given a Python object, returns a 32 character hash string to uniquely identify
    the content of the object. The object can be arbitrary nested (i.e. dictionary
    of dictionary of list etc), and hashing is applied recursively to uniquely
    identify the content.

    For dictionaries (at any level), the key order is ignored when hashing
    so that {"a":5, "b": 3, "c": 4} and {"b": 3, "a": 5, "c": 4} will both
    give rise to the same hash. Exception to this rule is when an OrderedDict
    is passed, in which case difference in key order is respected. To keep
    compatible with previous versions of Python and the assumed general
    intentions, key order will be ignored even in Python 3.7+ where the
    default dictionary is officially an ordered dictionary.

    Args:
        obj - A (potentially nested) Python object

    Returns:
        hash: str - a 32 charcter long hash string to uniquely identify the object.
    """
    hashed = hashlib.md5()

    if isinstance(obj, str):
        hashed.update(obj.encode())
    elif isinstance(obj, OrderedDict):
        for k, v in obj.items():
            hashed.update(str(k).encode())
            hashed.update(make_hash(v).encode())
    elif isinstance(obj, Mapping):
        for k in sorted(obj, key=str):
            hashed.update(str(k).encode())
            hashed.update(make_hash(obj[k]).encode())
    elif isinstance(obj, Iterable):
        for v in obj:
            hashed.update(make_hash(v).encode())
    else:
        hashed.update(str(obj).encode())

    return hashed.hexdigest()


def need_to_commit(repo, repo_name=""):
    changed_files = [item.a_path for item in repo.index.diff(None)]
    has_uncommited = bool(changed_files) or bool(repo.untracked_files)

    err_msg = []
    if has_uncommited:
        err_msg.append("\n{}".format(repo_name))
        if repo.untracked_files:
            for f in repo.untracked_files:
                err_msg.append("Untracked: \t" + f)
        if changed_files:
            for f in changed_files:
                err_msg.append("Changed: \t" + f)

    return "\n".join(err_msg)


def get_origin_url(g):
    for remote in g.remote(verbose=True).split("\n"):
        if remote.find("origin") + 1:
            origin_url = remote.split(" ")[0].split("origin\t")[-1]
            return origin_url
        else:
            warnings.warn("The repo does not have any remote url, named origin, specified.")


def check_repo_commit(repo_path):
    repo = Repo(path=repo_path)
    g = cmd.Git(repo_path)
    origin_url = get_origin_url(g)
    repo_name = origin_url.split("/")[-1].split(".")[0]
    err_msg = need_to_commit(repo, repo_name=repo_name)

    if err_msg:
        return "{}_error_msg".format(repo_name), err_msg

    else:
        sha1, branch = repo.head.commit.name_rev.split()
        commit_date = datetime.fromtimestamp(repo.head.commit.authored_date).strftime("%A %d. %B %Y %H:%M:%S")
        committer_name = repo.head.commit.committer.name
        committer_email = repo.head.commit.committer.email

        return (
            repo_name,
            {
                "sha1": sha1,
                "branch": branch,
                "commit_date": commit_date,
                "committer_name": committer_name,
                "committer_email": committer_email,
                "origin_url": origin_url,
            },
        )


def gitlog(repos=()):
    """
    A decorator on computed/imported tables.
    Monitors a list of repositories as pointed out by `repos` containing a list of paths to Git repositories. If any of these repositories
    contained uncommitted changes, the `populate` is interrupted.
    Otherwise, the state of commits associated with all repositoreis are summarized and stored in the associated entry in the GitLog part table.

    Example:

    @schema
    @gitlog(['/path/to/repo1', '/path/to/repo2'])
    class MyComputedTable(dj.Computed):
        ...

    """

    def gitlog_wrapper(cls):
        # if repos list is empty, skip the modification alltogether
        if len(repos) == 0:
            return cls

        class GitLog(dj.Part):
            definition = """
            ->master
            ---
            info :              longblob
            """

        def check_git(self):
            commits_info = {name: info for name, info in [check_repo_commit(repo) for repo in repos]}
            assert len(commits_info) == len(repos)

            if any(["error_msg" in name for name in commits_info.keys()]):
                err_msgs = ["You have uncommited changes."]
                err_msgs.extend([info for name, info in commits_info.items() if "error_msg" in name])
                err_msgs.append("\nPlease commit the changes before running populate.\n")
                raise RuntimeError("\n".join(err_msgs))

            return commits_info

        cls._base_populate = cls.populate
        cls._base_make = cls.make
        cls.check_git = check_git
        cls.GitLog = GitLog
        cls._commits_info = None

        def alt_populate(self, *args, **kwargs):
            # the commits info must be attached to the class
            # as table instance is NOT shared between populate and
            # make calls
            self.__class__._commits_info = self.check_git()
            ret = self._base_populate(*args, **kwargs)
            self.__class__._commits_info = None
            return ret

        def alt_make(self, key):
            ret = self._base_make(key)
            if self._commits_info is not None:
                # if there was some Git commit info
                entry = dict(key, info=self._commits_info)
                self.GitLog().insert1(entry)
            return ret

        cls.populate = alt_populate
        cls.make = alt_make

        return cls

    return gitlog_wrapper


def create_param_expansion(
    f_name,
    container_table,
    fn_field=None,
    config_field=None,
    resolver=None,
    suffix="Param",
    default_to_str=False,
):
    """
    Given a function name `f_name` as would be found in the `container_table` class, this will create
    a new DataJoint computed table subclass with the correct definition to expand blobs corresponding to
    the `f_name`'s arguments.

    The `container_table` must be a class (not an instance), and is also
    expected to implement `resolve_fn` method that can be used to resolve the name of the function to
    a specific function object. The name of the attributes for the function name and function argument
    object inside `container_table` is automatically inferred based on attribute names (i.e. ending with
    `_fn` and `_config`, repsectively).

    Alternatively, you can specifically supply a `resolver` function to resolve the function name,
    and also specify the name of the function and config object attributes via `fn_field` and `config_field`.
    Resolver functions are implemented in nnfabrik.builder and can be imported from there (`resolve_model`)

    The resulting computed table will have a name of the form `MyFunctionNameParam` for a function named
    `my_function_name`. In otherwords, the name is converted from snake_case to CamelCase and `suffix`
    (default to 'Param') is appended.
    """

    if fn_field is None:
        fn_field = next(v for v in container_table.heading.attributes.keys() if v.endswith("_fn"))

    if config_field is None:
        config_field = next(v for v in container_table.heading.attributes.keys() if v.endswith("_config"))

    resolver = resolver or (lambda x: container_table.resolve_fn(x))
    f = resolver(f_name)

    def_str = make_definition(f, default_to_str=default_to_str)

    class NewTable(dj.Computed):
        definition = """
        -> {}
        ---
        {}
        """.format(
            container_table.__name__, def_str
        )

        @property
        def key_source(self):
            return container_table & '{}="{}"'.format(fn_field, f_name)

        def make(self, key):
            entries = (container_table & key).fetch1(config_field)
            entries = cleanup_numpy_scalar(entries)
            key = dict(key, **entries)
            if default_to_str:
                for k, v in key.items():
                    if type(v) in [list, tuple]:
                        key[k] = str(v)
            self.insert1(key, ignore_extra_fields=True)

    NewTable.__name__ = to_camel_case(f.__name__) + suffix
    return NewTable


def make_definition(f, exclude=("model", "dataloaders", "seed"), default_to_str=False):
    """
    Given a function `f`, creates a table definition string to house all arguments. The types
    of the arguments are inferred from (1) type annotation and (2) type of the default value if present,
    in that order. If type cannot be inferred, defaults to `longblob`.
    Arguments matching values in the exclude list will not be included in the definition.
    """
    type_lut = {
        int: "int",
        float: "float",
        str: "varchar(255)",
        date: "DATE",
        datetime: "DATETIME",
        object: "longblob",
        bool: "bool",
    }
    argspec = inspect.getfullargspec(f)
    total_def = []
    def_lut = {}
    if argspec.defaults is not None:
        def_lut = {k: d for k, d in zip(argspec.args[::-1], argspec.defaults[::-1])}

    for v in argspec.args:
        # skip arguments found in the exclude list
        if v in exclude:
            continue
        if v in argspec.annotations:
            t = argspec.annotations[v]
            if not t in [str, int, float, bool]:
                if default_to_str:
                    t = str
                else:
                    t = object
        elif v in def_lut:
            t = type(def_lut[v])
            if not t in [str, int, float, bool]:
                if default_to_str:
                    t = str
                else:
                    t = object
        else:
            t = object
        field = type_lut.get(t, "longblob")  # default to longblob if no match found

        # if boolean field, turn default value into an integer
        if field == "bool" and v in def_lut:
            def_lut[v] = int(def_lut[v])

        if v in def_lut:
            default_clause = "={!r}".format(def_lut[v]) if def_lut[v] is not None else "=NULL"
        else:
            default_clause = ""
        spec_str = "{}{}: {}   # autogenerated column - {}".format(v, default_clause, field, field)
        total_def.append(spec_str)
    return "\n".join(total_def)


class CustomSchema(Schema):
    def __call__(self, cls, *, context=None):
        context = context or self.context or inspect.currentframe().f_back.f_locals
        # Process all part tables and replace with a subclass
        for attr in dir(cls):
            if attr[0].isupper():
                part = getattr(cls, attr)
                if inspect.isclass(part) and issubclass(part, dj.Part):

                    class WrappedPartTable(part):
                        pass

                    WrappedPartTable.__name__ = attr
                    setattr(cls, attr, WrappedPartTable)
        return super().__call__(cls, context=context)



from typing import Dict

import numpy as np
from torch.utils.data import DataLoader
from torchvision import datasets, transforms


def mnist_dataset_fn(seed: int, **config) -> Dict:
    """
    Returns data loaders for the given config
    Args:
        seed: random seed that will make shuffling and other random operations deterministic
    Returns:
        data_loaders: containing "train", "validation" and "test" data loaders
    """
    np.random.seed(seed)

    transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,)),  # (mean,std) of MNIST train set
        ]
    )
    train_dataset = datasets.MNIST("../data", train=True, download=True, transform=transform)
    validation_dataset = datasets.MNIST(
        "../data", train=False, download=True, transform=transform
    )  # for simplicity, we use the test set for validation
    test_dataset = datasets.MNIST("../data", train=False, download=True, transform=transform)
    batch_size = config.get("batch_size", 64)
    return {
        "train": DataLoader(train_dataset, batch_size=batch_size),
        "validation": DataLoader(validation_dataset, batch_size=batch_size),
        "test": DataLoader(test_dataset, batch_size=batch_size),
    }



# helper functions concerning the ANN architecture

import torch
from torch import nn

import numpy as np
import random


def get_io_dims(data_loader):
    """
    Returns the shape of the dataset for each item within an entry returned by the `data_loader`
    The DataLoader object must return either a namedtuple, dictionary or a plain tuple.
    If `data_loader` entry is a namedtuple or a dictionary, a dictionary with the same keys as the
    namedtuple/dict item is returned, where values are the shape of the entry. Otherwise, a tuple of
    shape information is returned.

    Note that the first dimension is always the batch dim with size depending on the data_loader configuration.

    Args:
        data_loader (torch.DataLoader): is expected to be a pytorch Dataloader object returning
            either a namedtuple, dictionary, or a plain tuple.
    Returns:
        dict or tuple: If data_loader element is either namedtuple or dictionary, a ditionary
            of shape information, keyed for each entry of dataset is returned. Otherwise, a tuple
            of shape information is returned. The first dimension is always the batch dim
            with size depending on the data_loader configuration.
    """
    items = next(iter(data_loader))
    if hasattr(items, "_asdict"):  # if it's a named tuple
        items = items._asdict()

    if hasattr(items, "items"):  # if dict like
        return {k: v.shape for k, v in items.items()}
    else:
        return (v.shape for v in items)


def get_dims_for_loader_dict(dataloaders):
    """
    Given a dictionary of DataLoaders, returns a dictionary with same keys as the
    input and shape information (as returned by `get_io_dims`) on each keyed DataLoader.

    Args:
        dataloaders (dict of DataLoader): Dictionary of dataloaders.

    Returns:
        dict: A dict containing the result of calling `get_io_dims` for each entry of the input dict
    """
    return {k: get_io_dims(v) for k, v in dataloaders.items()}


def set_random_seed(seed: int, deterministic: bool = True):
    """
    Set random generator seed for Python interpreter, NumPy and PyTorch. When setting the seed for PyTorch,
    if CUDA device is available, manual seed for CUDA will also be set. Finally, if `deterministic=True`,
    and CUDA device is available, PyTorch CUDNN backend will be configured to `benchmark=False` and `deterministic=True`
    to yield as deterministic result as possible. For more details, refer to
    PyTorch documentation on reproducibility: https://pytorch.org/docs/stable/notes/randomness.html

    Beware that the seed setting is a "best effort" towards deterministic run. However, as detailed in the above documentation,
    there are certain PyTorch CUDA opertaions that are inherently non-deterministic, and there is no simple way to control for them.
    Thus, it is best to assume that when CUDA is utilized, operation of the PyTorch module will not be deterministic and thus
    not completely reproducible.

    Args:
        seed (int): seed value to be set
        deterministic (bool, optional): If True, CUDNN backend (if available) is set to be deterministic. Defaults to True. Note that if set
            to False, the CUDNN properties remain untouched and it NOT explicitly set to False.
    """
    random.seed(seed)
    np.random.seed(seed)
    if deterministic:
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
    torch.manual_seed(seed)  # this sets both CPU and CUDA seeds for PyTorch


def move_to_device(model, gpu=True, multi_gpu=True):
    """
    Moves given model to GPU(s) if they are available
    :param model: (torch.nn.Module) model to move
    :param gpu: (bool) if True attempt to move to GPU
    :param multi_gpu: (bool) if True attempt to use multi-GPU
    :return: torch.nn.Module, str
    """
    device = "cuda" if torch.cuda.is_available() and gpu else "cpu"
    if multi_gpu and torch.cuda.device_count() > 1:
        print("Using ", torch.cuda.device_count(), "GPUs")
        model = nn.DataParallel(model)
    model = model.to(device)
    return model, device


def find_prefix(keys: list, p_agree: float = 0.66, separator=".") -> (list, int):
    """
    Finds common prefix among state_dict keys
    :param keys: list of strings to find a common prefix
    :param p_agree: percentage of keys that should agree for a valid prefix
    :param separator: string that separates keys into substrings, e.g. "model.conv1.bias"
    :return: (prefix, end index of prefix)
    """
    keys = [k.split(separator) for k in keys]
    p_len = 0
    common_prefix = ""
    prefs = {"": len(keys)}
    while True:
        sorted_prefs = sorted(prefs.items(), key=lambda x: x[1], reverse=True)
        # check if largest count is above threshold
        if sorted_prefs[0][1] < p_agree * len(keys):
            break
        common_prefix = sorted_prefs[0][0]  # save prefix

        p_len += 1
        prefs = {}
        for key in keys:
            if p_len == len(key):  # prefix cannot be an entire key
                continue
            p_str = ".".join(key[:p_len])
            prefs[p_str] = prefs.get(p_str, 0) + 1
    return common_prefix, p_len - 1


def load_state_dict(
    model,
    state_dict: dict,
    ignore_missing: bool = False,
    ignore_unused: bool = False,
    match_names: bool = False,
    ignore_dim_mismatch: bool = False,
    prefix_agreement: float = 0.66,
):
    """
    Loads given state_dict into model, but allows for some more flexible loading.

    :param model: nn.Module object
    :param state_dict: dictionary containing a whole state of the module (result of `some_model.state_dict()`)
    :param ignore_missing: if True ignores entries present in model but not in `state_dict`
    :param match_names: if True tries to match names in `state_dict` and `model.state_dict()`
                        by finding and removing a common prefix from the keys in each dict
    :param ignore_dim_mismatch: if True ignores parameters in `state_dict` that don't fit the shape in `model`
    """

    model_dict = model.state_dict()
    # 0. Try to match names by adding or removing prefix:
    if match_names:
        # find prefix keys of each state dict:
        s_pref, s_idx = find_prefix(list(state_dict.keys()), p_agree=prefix_agreement)
        m_pref, m_idx = find_prefix(list(model_dict.keys()), p_agree=prefix_agreement)
        # switch prefixes:
        stripped_state_dict = {}
        for k, v in state_dict.items():
            if k.split(".")[:s_idx] == s_pref.split("."):
                stripped_key = ".".join(k.split(".")[s_idx:])
            else:
                stripped_key = k
            new_key = m_pref + "." + stripped_key if m_pref else stripped_key
            stripped_state_dict[new_key] = v
        state_dict = stripped_state_dict

    # 1. filter out missing keys
    filtered_state_dict = {k: v for k, v in state_dict.items() if k in model_dict}
    unused = set(state_dict.keys()) - set(filtered_state_dict.keys())
    if unused and ignore_unused:
        print("Ignored unnecessary keys in pretrained dict:\n" + "\n".join(unused))
    elif unused:
        raise RuntimeError("Error in loading state_dict: Unused keys:\n" + "\n".join(unused))
    missing = set(model_dict.keys()) - set(filtered_state_dict.keys())
    if missing and ignore_missing:
        print("Ignored Missing keys:\n" + "\n".join(missing))
    elif missing:
        raise RuntimeError("Error in loading state_dict: Missing keys:\n" + "\n".join(missing))

    # 2. overwrite entries in the existing state dict
    updated_model_dict = {}
    for k, v in filtered_state_dict.items():
        if v.shape != model_dict[k].shape:
            if ignore_dim_mismatch:
                print("Ignored shape-mismatched parameter:", k)
                continue
            else:
                raise RuntimeError("Error in loading state_dict: Shape-mismatch for key {}".format(k))
        updated_model_dict[k] = v

    # 3. load the new state dict
    model.load_state_dict(updated_model_dict, strict=(not ignore_missing))



import tempfile
import torch
import os
import warnings
import datajoint as dj
from datajoint.fetch import DataJointError
from ..builder import get_all_parts, get_model, get_trainer
from ..utility.dj_helpers import make_hash
from .utility import find_object


class TrainedModelBase(dj.Computed):
    """
    Base class for defining TrainedModel table used to tigger training of models in nnfabrik.

    To use this class, define a new class inheriting from this base class, and decorate with your own
    schema. Furthermore, you have to do one of the following for the class to be functional:
    * Set the class property `nnfabrik` to point to a module or a dictionary context that contains classes
        for tables corresponding to `Fabrikant`, `Seed`, `Dataset`, `Model`, and `Trainer`. Most commonly, you
        would want to simply pass the resulting module object from `my_nnfabrik` output.
    * Set the class property `nnfabrik` to "core" -- this will then make this table refer to
        `Fabrikant`, `Seed`, `Dataset`, `Model`, and `Trainer` as found inside `main` module directly. Note that
        this will therefore depend on the shared "core" tables of nnfabrik.
    * Set the values of the following class properties to individually specify the DataJoint table to use:
        `user_table`, `seed_table`, `dataset_table`, `model_table` and `trainer_table` to specify equivalent
        of `Fabrikant`, `Seed`, `Dataset`, `Model`, and `Trainer`, respectively. You could also set the
        value of `nnfabrik` to a module or "core" as stated above, and specifically override a target table
        via setting one of the table class property as well.
    """

    database = ""  # hack to suppress DJ error

    nnfabrik = None

    @property
    def model_table(self):
        return find_object(self.nnfabrik, "Model")

    @property
    def dataset_table(self):
        return find_object(self.nnfabrik, "Dataset")

    @property
    def trainer_table(self):
        return find_object(self.nnfabrik, "Trainer")

    @property
    def seed_table(self):
        return find_object(self.nnfabrik, "Seed")

    @property
    def user_table(self):
        return find_object(self.nnfabrik, "Fabrikant", "user_table")

    @property
    def data_info_table(self):
        return find_object(self.nnfabrik, "DataInfo", "data_info_table")

    # storage for the ModelStorage table
    storage = "minio"

    # delimitter to use when concatenating comments from model, dataset, and trainer tables
    comment_delimitter = "."

    # table level comment
    table_comment = "Trained models"

    @property
    def definition(self):
        definition = """
        # {table_comment}
        -> self().model_table
        -> self().dataset_table
        -> self().trainer_table
        -> self().seed_table
        ---
        comment='':                        varchar(768) # short description 
        score:                             float        # loss
        output:                            longblob     # trainer object's output
        ->[nullable] self().user_table
        trainedmodel_ts=CURRENT_TIMESTAMP: timestamp    # UTZ timestamp at time of insertion
        """.format(
            table_comment=self.table_comment
        )
        return definition

    class ModelStorage(dj.Part):
        @property
        def definition(self):
            definition = """
            # Contains the models state dict, stored externally.
            -> master
            ---
            model_state:            attach@{storage}
            """.format(
                storage=self._master.storage
            )
            return definition

    def get_full_config(self, key=None, include_state_dict=True, include_trainer=True):
        """
        Returns the full configuration dictionary needed to build all components of the network
        training including dataset, model and trainer. The returned dictionary is designed to be
        passed (with dictionary expansion) into the get_all_parts function provided in builder.py.

        Args:
            key - specific key against which to retrieve all configuration. The key must restrict all component
                  tables into a single entry. If None, will assume that this table is already restricted and
                  will obtain an existing single entry.
            include_state_dict (bool) : If True, and if key refers to a model already trained with a corresponding entry in self.ModelStorage,
                  the state_dict of the trained model is retrieved and returned
            include_trainer (bool): If False, then trainer configuration is skipped. Usually desirable when you want to simply retrieve trained model.
        """
        if key is None:
            key = self.fetch1("KEY")

        model_fn, model_config = (self.model_table & key).fn_config
        dataset_fn, dataset_config = (self.dataset_table & key).fn_config

        ret = dict(
            model_fn=model_fn,
            model_config=model_config,
            dataset_fn=dataset_fn,
            dataset_config=dataset_config,
        )

        if include_trainer:
            trainer_fn, trainer_config = (self.trainer_table & key).fn_config
            ret["trainer_fn"] = trainer_fn
            ret["trainer_config"] = trainer_config

        # if trained model exist and include_state_dict is True
        if include_state_dict and (self.ModelStorage & key):
            with tempfile.TemporaryDirectory() as temp_dir:
                state_dict_path = (self.ModelStorage & key).fetch1("model_state", download_path=temp_dir)
                ret["state_dict"] = torch.load(state_dict_path)

        return ret

    def load_model(
        self,
        key=None,
        include_dataloader=True,
        include_trainer=False,
        include_state_dict=True,
        seed: int = None,
    ):
        """
        Load a single entry of the model. If state_dict is available, the model will be loaded with state_dict as well.
        By default the trainer is skipped. Set `include_trainer=True` to also retrieve the trainer function
        as the third return argument.

        Args:
            key - specific key against which to retrieve the model. The key must restrict all component
                  tables into a single entry. If None, will assume that this table is already restricted and
                  will obtain an existing single entry.
            include_dataloader - if True, builds the dataloaer and the model, and returns both.
                                 if False, tries to build the model without requiring dataloader.
                                    Returns the model only when set to False.
            include_trainer - If False (default), will not load or return the trainer.
            include_state_dict - If True, the model is loaded with state_dict if key corresponds to a trained entry.
            seed - Optional seed. If not given and a corresponding entry exists in self.seed_table, seed is taken from there

        Returns
            dataloaders - Loaded dictionary (train, test, validation) of dictionary (data_key) of dataloaders
            model - Loaded model. If key corresponded to an existing entry, it would have also loaded the
                    state_dict unless load_state_dict=False
            trainer - Loaded trainer function. This is not returned if include_trainer=False.
        """
        if key is None:
            key = self.fetch1("KEY")

        # if no explicit seed is provided and there is already a corresponding entry in the seed_table
        # use that seed value
        if seed is None and len(self.seed_table & key) == 1:
            seed = (self.seed_table & key).fetch1("seed")

        config_dict = self.get_full_config(key, include_trainer=include_trainer, include_state_dict=include_state_dict)

        if not include_dataloader:
            try:
                data_info = (self.data_info_table & key).fetch1("data_info")
                model_config_dict = dict(
                    model_fn=config_dict["model_fn"],
                    model_config=config_dict["model_config"],
                    data_info=data_info,
                    seed=seed,
                    state_dict=config_dict.get("state_dict", None),
                    strict=False,
                )

                net = get_model(**model_config_dict)
                return (
                    (
                        net,
                        get_trainer(config_dict["trainer_fn"], config_dict["trainer_config"]),
                    )
                    if include_trainer
                    else net
                )

            except (TypeError, AttributeError, DataJointError):
                warnings.warn(
                    "Model could not be built without the dataloader. Dataloader will be built in order to create the model. "
                    "Make sure to have an The 'model_fn' also has to be able to"
                    "accept 'data_info' as an input arg, and use that over the dataloader to build the model."
                )

            ret = get_all_parts(**config_dict, seed=seed)
            return ret[1:] if include_trainer else ret[1]

        return get_all_parts(**config_dict, seed=seed)

    def call_back(self, uid=None, epoch=None, model=None, info=None):
        """
        Override this implementation to get periodic calls during the training
        by the trainer.

        Args:
            uid - Unique identifier for the trained model entry
            epoch - the iteration count
            model - current model under training
            info - Additional information provided by the trainer
        """
        pass

    def make(self, key):
        """
        Given key specifying configuration for dataloaders, model and trainer,
        trains the model and saves the trained model.
        """
        # lookup the fabrikant corresponding to the current DJ user
        fabrikant_name = self.user_table.get_current_user()
        seed = (self.seed_table & key).fetch1("seed")

        # load everything
        dataloaders, model, trainer = self.load_model(key, include_trainer=True, include_state_dict=False, seed=seed)

        # define callback with pinging
        def call_back(**kwargs):
            self.connection.ping()
            self.call_back(**kwargs)

        # model training
        score, output, model_state = trainer(model=model, dataloaders=dataloaders, seed=seed, uid=key, cb=call_back)

        # save resulting model_state into a temporary file to be attached
        with tempfile.TemporaryDirectory() as temp_dir:
            filename = make_hash(key) + ".pth.tar"
            filepath = os.path.join(temp_dir, filename)
            torch.save(model_state, filepath)

            key["score"] = score
            key["output"] = output
            key["fabrikant_name"] = fabrikant_name
            comments = []
            comments.append((self.trainer_table & key).fetch1("trainer_comment"))
            comments.append((self.model_table & key).fetch1("model_comment"))
            comments.append((self.dataset_table & key).fetch1("dataset_comment"))
            key["comment"] = self.comment_delimitter.join(comments)
            self.insert1(key)

            key["model_state"] = filepath

            self.ModelStorage.insert1(key, ignore_extra_fields=True)


class DataInfoBase(dj.Computed):
    """
    Inherit from this class and decorate with your own schema to create a functional
    DataInfo table.

    Furthermore, you have to do one of the following for the class to be functional:
    * Set the class property `nnfabrik` to point to a module or a dictionary context that contains classes
        for tables corresponding to `Fabrikant` and `Dataset`. Most commonly, you
        would want to simply pass the resulting module object from `my_nnfabrik` output.
    * Set the class property `nnfabrik` to "core" -- this will then make this table refer to
        `Fabrikant` and `Dataset` as found inside `main` module directly. Note that
        this will therefore depend on the shared "core" tables of nnfabrik.
    * Set the values of the following class properties to individually specify the DataJoint table to use:
        `user_table` and `dataset_table` to specify equivalent
        of `Fabrikant` and `Dataset` respectively. You could also set the
        value of `nnfabrik` to a module or "core" as stated above, and specifically override a target table
        via setting one of the table class property as well.
    """

    nnfabrik = None

    @property
    def dataset_table(self):
        return find_object(self.nnfabrik, "Dataset")

    @property
    def user_table(self):
        return find_object(self.nnfabrik, "Fabrikant", "user_table")

    # table level comment
    table_comment = "Table containing information about i/o dimensions and statistics, per data_key in dataset"

    @property
    def definition(self):
        definition = """
            # {table_comment}
            -> self().dataset_table
            ---
            data_info:                     longblob     # Dictionary of data_keys and i/o information

            ->[nullable] self().user_table
            datainfo_ts=CURRENT_TIMESTAMP: timestamp    # UTZ timestamp at time of insertion
            """.format(
            table_comment=self.table_comment
        )
        return definition

    def make(self, key):
        """
        Given a dataset from nnfabrik, extracts the necessary information for building a model in nnfabrik.
        'data_info' is expected to be a dictionary of dictionaries, similar to the dataloaders object.
        For example:
            data_info = {
                        'data_key_0': dict(input_dimensions=[N,c,h,w, ...],
                                           input_channels=[c],
                                           output_dimension=[o, ...],
                                           img_mean=mean_train_images,
                                           img_std=std_train_images),
                        'data_key_1':  ...
                        }
        """
        dataset_fn, dataset_config = (self.dataset_table & key).fn_config
        data_info = dataset_fn(**dataset_config, return_data_info=True)

        fabrikant_name = self.user_table.get_current_user()

        key["fabrikant_name"] = fabrikant_name
        key["data_info"] = data_info
        self.insert1(key)



from importlib import import_module
import numpy as np
from ..utility.dj_helpers import make_hash, cleanup_numpy_scalar


def split_module_name(abs_class_name):
    abs_module_path = ".".join(abs_class_name.split(".")[:-1])
    class_name = abs_class_name.split(".")[-1]
    return (abs_module_path, class_name)


def dynamic_import(abs_module_path, class_name):
    module_object = import_module(abs_module_path)
    target_class = getattr(module_object, class_name)
    return target_class


class FabrikCache:
    def __init__(self, base_table, cache_size_limit=10):
        self.base_table = base_table
        self.cache_size_limit = cache_size_limit
        self.cache = dict()
        if hasattr(self.base_table, "load_model"):
            self.load_function = self.base_table().load_model
        elif hasattr(self.base_table, "get_dataloader"):
            self.load_function = self.base_table().get_dataloader
        elif hasattr(self.base_table, "build_model"):
            self.load_function = self.base_table().build_model
        else:
            raise ValueError("Base table needs to have a 'load_model', 'get_dataloader', or 'build_model' method")

    def load(self, key, **kwargs):
        if self.cache_size_limit == 0:
            return self._load_model(key, **kwargs)
        if not self._is_cached(key):
            self._cache_model(key, **kwargs)
        return self._get_cached_model(key)

    def _load_model(self, key, **kwargs):
        return self.load_function(key=key, **kwargs)

    def _is_cached(self, key):
        if self._hash_trained_model_key(key) in self.cache:
            return True
        return False

    def _cache_model(self, key, **kwargs):
        """Caches a model and makes sure the cache is not bigger than the specified limit."""
        self.cache[self._hash_trained_model_key(key)] = self._load_model(key, **kwargs)
        if len(self.cache) > self.cache_size_limit:
            del self.cache[list(self.cache)[0]]

    def _get_cached_model(self, key):
        return self.cache[self._hash_trained_model_key(key)]

    def _hash_trained_model_key(self, key):
        """Creates a hash from the part of the key corresponding to the primary key of the trained model table."""
        return make_hash({k: key[k] for k in self.base_table().primary_key})



import numpy as np
from ax.service.managed_loop import optimize
from .nnf_helper import split_module_name, dynamic_import
from nnfabrik.main import *
import datajoint as dj


class Bayesian:
    """
    A hyperparameter optimization tool based on Facebook Ax (https://ax.dev/), integrated with nnfabrik.
    This tool, iteratively, optimizes for hyperparameters that improve a specific score representing
    model performance (the same score used in TrainedModel table). In every iteration (after every training),
    it automatically adds an entry to the corresponding tables, and populated the trained model table (i.e.
    trains the model) for that specific entry.

    Args:
        dataset_fn (str): name of the dataset function
        dataset_config (dict): dictionary of arguments for dataset function that are fixed
        dataset_config_auto (dict): dictionary of arguments for dataset function that are to be optimized
        model_fn (str): name of the model function
        model_config (dict): dictionary of arguments for model function that are fixed
        model_config_auto (dict): dictionary of arguments for model function that are to be optimized
        trainer_fn (str): name of the trainer function
        trainer_config (dict): dictionary of arguments for trainer function that are fixed
        trainer_config_auto (dict): dictionary of arguments for trainer function that are to be optimized
        architect (str): Name of the contributor that added this entry
        trained_model_table (str): name (importable) of the trained_model_table
        total_trials (int, optional): Number of experiments (i.e. training) to run. Defaults to 5.
        arms_per_trial (int, optional): Number of different configurations used for training (for more details check https://ax.dev/docs/glossary.html#trial). Defaults to 1.
        comment (str, optional): Comments about this optimization round. It will be used to fill up the comment entry of dataset, model, and trainer table. Defaults to "Bayesian optimization of Hyper params.".
    """

    def __init__(
        self,
        dataset_fn,
        dataset_config,
        dataset_config_auto,
        model_fn,
        model_config,
        model_config_auto,
        trainer_fn,
        trainer_config,
        trainer_config_auto,
        architect,
        trained_model_table,
        total_trials=5,
        arms_per_trial=1,
        comment="Bayesian optimization of Hyper params.",
    ):

        self.fns = dict(dataset=dataset_fn, model=model_fn, trainer=trainer_fn)
        self.fixed_params = self.get_fixed_params(dataset_config, model_config, trainer_config)
        self.auto_params = self.get_auto_params(dataset_config_auto, model_config_auto, trainer_config_auto)
        self.architect = architect
        self.total_trials = total_trials
        self.arms_per_trial = arms_per_trial
        self.comment = comment

        # import TrainedModel definition
        module_path, class_name = split_module_name(trained_model_table)
        self.trained_model_table = dynamic_import(module_path, class_name)

    @staticmethod
    def get_fixed_params(dataset_config, model_config, trainer_config):
        """
        Returns a single dictionary including the fixed parameters for dataset, model, and trainer.

        Args:
            dataset_config (dict): dictionary of arguments for dataset function that are fixed
            model_config (dict): dictionary of arguments for model function that are fixed
            trainer_config (dict): dictionary of arguments for trainer function that are fixed

        Returns:
            dict: A dictionary of dictionaries where keys are dataset, model, and trainer and the values
            are the corresponding dictionary of fixed arguments.
        """
        return dict(dataset=dataset_config, model=model_config, trainer=trainer_config)

    @staticmethod
    def get_auto_params(dataset_config_auto, model_config_auto, trainer_config_auto):
        """
        Changes the parameters to be optimized to a ax-friendly format, i.e. a list of dictionaries where
        each entry of the list corresponds to a single parameter.

        Ax requires a list of parameters (to be optimized) including the name and other specifications.
        Here we provide that list while keeping the arguments still separated by adding "dataset",
        "model", or "trainer" to the beginning of the parameter name.

        Args:
            dataset_config_auto (dict): dictionary of arguments for dataset function that are to be optimized
            model_config_auto (dict): dictionary of arguments for model function that are to be optimized
            trainer_config_auto (dict): dictionary of arguments for trainer function that are to be optimized

        Returns:
            list: list of dictionaries where each dictionary specifies a single parameter to be optimized.
        """
        dataset_params = []
        for k, v in dataset_config_auto.items():
            dd = {"name": "dataset.{}".format(k)}
            dd.update(v)
            dataset_params.append(dd)

        model_params = []
        for k, v in model_config_auto.items():
            dd = {"name": "model.{}".format(k)}
            dd.update(v)
            model_params.append(dd)

        trainer_params = []
        for k, v in trainer_config_auto.items():
            dd = {"name": "trainer.{}".format(k)}
            dd.update(v)
            trainer_params.append(dd)

        return dataset_params + model_params + trainer_params

    @staticmethod
    def _combine_params(auto_params, fixed_params):
        """
        Combining the auto (to-be-optimized) and fixed parameters (to have a single object representing all the arguments
        used for a specific function)

        Args:
            auto_params (dict): dictionary of to-be-optimized parameters, i.e. A dictionary of dictionaries where keys are
            dataset, model, and trainer and the values are the corresponding dictionary of to-be-optimized arguments.
            fixed_params (dict): dictionary of fixed parameters, i.e. A dictionary of dictionaries where keys are dataset,
            model, and trainer and the values are the corresponding dictionary of fixed arguments.

        Returns:
            dict: dictionary of parameters (fixed and to-be-optimized), i.e. A dictionary of dictionaries where keys are
            dataset, model, and trainer and the values are the corresponding dictionary of arguments.
        """
        keys = ["dataset", "model", "trainer"]
        params = {}
        for key in keys:
            params[key] = fixed_params[key]
            params[key].update(auto_params[key])

        return {key: params[key] for key in keys}

    @staticmethod
    def _split_config(params):
        """
        Reverses the operation of `get_auto_params` (from ax-friendly format to a dictionary of dictionaries where keys are
        dataset, model, and trainer and the values are a dictionary of the corresponding arguments)

        Args:
            params (dict): dictionary of dictionaries where each dictionary specifies a single parameter to be optimized.

        Returns:
            dict: A dictionary of dictionaries where keys are dataset, model, and trainer and the values are the
            corresponding dictionary of to-be-optimized arguments.
        """
        config = dict(dataset={}, model={}, trainer={}, others={})
        for k, v in params.items():
            config[k.split(".")[0]][k.split(".")[1]] = v

        return config

    def train_evaluate(self, auto_params):
        """
        For a given set of parameters, add an entry to the corresponding tables, and populated the trained model
        table for that specific entry.

        Args:
            auto_params (dict): dictionary of dictionaries where each dictionary specifies a single parameter to be optimized.

        Returns:
            float: the score of the trained model for the specific entry in trained model table
        """
        config = self._combine_params(self._split_config(auto_params), self.fixed_params)

        # insert the stuff into their corresponding tables
        dataset_hash = make_hash(config["dataset"])
        entry_exists = {
            "dataset_fn": "{}".format(self.fns["dataset"])
        } in self.trained_model_table.dataset_table() and {
            "dataset_hash": "{}".format(dataset_hash)
        } in self.trained_model_table.dataset_table()
        if not entry_exists:
            self.trained_model_table.dataset_table().add_entry(
                self.fns["dataset"],
                config["dataset"],
                dataset_fabrikant=self.architect,
                dataset_comment=self.comment,
            )

        model_hash = make_hash(config["model"])
        entry_exists = {"model_fn": "{}".format(self.fns["model"])} in self.trained_model_table.model_table() and {
            "model_hash": "{}".format(model_hash)
        } in self.trained_model_table.model_table()
        if not entry_exists:
            self.trained_model_table.model_table().add_entry(
                self.fns["model"],
                config["model"],
                model_fabrikant=self.architect,
                model_comment=self.comment,
            )

        trainer_hash = make_hash(config["trainer"])
        entry_exists = {
            "trainer_fn": "{}".format(self.fns["trainer"])
        } in self.trained_model_table.trainer_table() and {
            "trainer_hash": "{}".format(trainer_hash)
        } in self.trained_model_table.trainer_table()
        if not entry_exists:
            self.trained_model_table.trainer_table().add_entry(
                self.fns["trainer"],
                config["trainer"],
                trainer_fabrikant=self.architect,
                trainer_comment=self.comment,
            )

        # get the primary key values for all those entries
        restriction = (
            'dataset_fn in ("{}")'.format(self.fns["dataset"]),
            'dataset_hash in ("{}")'.format(dataset_hash),
            'model_fn in ("{}")'.format(self.fns["model"]),
            'model_hash in ("{}")'.format(model_hash),
            'trainer_fn in ("{}")'.format(self.fns["trainer"]),
            'trainer_hash in ("{}")'.format(trainer_hash),
        )

        # populate the table for those primary keys
        self.trained_model_table().populate(*restriction)

        # get the score of the model for this specific set of hyperparameters
        score = (self.trained_model_table() & dj.AndList(restriction)).fetch("score")[0]

        return score

    def run(self):
        """
        Runs Bayesian optimization.

        Returns:
            tuple: The returned values are similar to that of Ax (refer to https://ax.dev/docs/api.html)
        """
        best_parameters, values, experiment, model = optimize(
            parameters=self.auto_params,
            evaluation_function=self.train_evaluate,
            objective_name="val_corr",
            minimize=False,
            total_trials=self.total_trials,
            arms_per_trial=self.arms_per_trial,
        )

        return self._split_config(best_parameters), values, experiment, model


class Random:
    """
    Random hyperparameter search, integrated with nnfabrik.
    Similar to Bayesian optimization tool, but instead of optimizing for hyperparameters to maximize a score,
    in every iteration (after every training), it randomly samples new value for the specified parameters, adds an
    entry to the corresponding tables, and populated the trained model table (i.e. trains the model) for that specific entry.

    Args:
        dataset_fn (str): name of the dataset function
        dataset_config (dict): dictionary of arguments for dataset function that are fixed
        dataset_config_auto (dict): dictionary of arguments for dataset function that are to be randomly sampled
        model_fn (str): name of the model function
        model_config (dict): dictionary of arguments for model function that are fixed
        model_config_auto (dict): dictionary of arguments for model function that are to be randomly sampled
        trainer_fn (str): name of the trainer function
        trainer_config (dict): dictionary of arguments for trainer function that are fixed
        trainer_config_auto (dict): dictionary of arguments for trainer function that are to be randomly sampled
        architect (str): Name of the contributor that added this entry
        trained_model_table (str): name (importable) of the trained_model_table
        total_trials (int, optional): Number of experiments (i.e. training) to run. Defaults to 5.
        comment (str, optional): Comments about this optimization round. It will be used to fill up the comment entry of dataset, model, and trainer table. Defaults to "Bayesian optimization of Hyper params.".
    """

    def __init__(
        self,
        dataset_fn,
        dataset_config,
        dataset_config_auto,
        model_fn,
        model_config,
        model_config_auto,
        trainer_fn,
        trainer_config,
        trainer_config_auto,
        architect,
        trained_model_table,
        total_trials=5,
        comment="Random search for hyper params.",
    ):

        self.fns = dict(dataset=dataset_fn, model=model_fn, trainer=trainer_fn)
        self.fixed_params = self.get_fixed_params(dataset_config, model_config, trainer_config)
        self.auto_params = self.get_auto_params(dataset_config_auto, model_config_auto, trainer_config_auto)
        self.architect = architect
        self.total_trials = total_trials
        self.comment = comment

        # import TrainedModel definition
        module_path, class_name = split_module_name(trained_model_table)
        self.trained_model_table = dynamic_import(module_path, class_name)

    @staticmethod
    def get_fixed_params(dataset_config, model_config, trainer_config):
        """
        Returs a single dictionary including the fixed parameters for dataset, model, and trainer.

        Args:
            dataset_config (dict): dictionary of arguments for dataset function that are fixed
            model_config (dict): dictionary of arguments for model function that are fixed
            trainer_config (dict): dictionary of arguments for trainer function that are fixed

        Returns:
            dict: A dictionary of dictionaries where keys are dataset, model, and trainer and the values are the corresponding
            dictionary of fixed arguments.
        """
        return dict(dataset=dataset_config, model=model_config, trainer=trainer_config)

    @staticmethod
    def get_auto_params(dataset_config_auto, model_config_auto, trainer_config_auto):
        """
        Returns the parameters, which are to be randomly sampled, in a list.
        Here we followed the same convention as in the Bayesian class, to have the API as similar as possible.

        Args:
            dataset_config_auto (dict): dictionary of arguments for dataset function that are to be randomly sampled
            model_config_auto (dict): dictionary of arguments for model function that are to be randomly sampled
            trainer_config_auto (dict): dictionary of arguments for trainer function that are to be randomly sampled

        Returns:
            list: list of dictionaries where each dictionary specifies a single parameter to be randomly sampled.
        """
        dataset_params = []
        for k, v in dataset_config_auto.items():
            dd = {"name": "dataset.{}".format(k)}
            dd.update(v)
            dataset_params.append(dd)

        model_params = []
        for k, v in model_config_auto.items():
            dd = {"name": "model.{}".format(k)}
            dd.update(v)
            model_params.append(dd)

        trainer_params = []
        for k, v in trainer_config_auto.items():
            dd = {"name": "trainer.{}".format(k)}
            dd.update(v)
            trainer_params.append(dd)

        return dataset_params + model_params + trainer_params

    @staticmethod
    def _combine_params(auto_params, fixed_params):
        """
        Combining the auto and fixed parameters (to have a single object representing all the arguments used for a specific function)

        Args:
            auto_params (dict): dictionary of to-be-sampled parameters, i.e. A dictionary of dictionaries where keys are dataset,
            model, and trainer and the values are the corresponding dictionary of to-be-sampled arguments.
            fixed_params (dict): dictionary of fixed parameters, i.e. A dictionary of dictionaries where keys are dataset, model,
            and trainer and the values are the corresponding dictionary of fixed arguments.

        Returns:
            dict: dictionary of parameters (fixed and to-be-sampled), i.e. A dictionary of dictionaries where keys are dataset,
            model, and trainer and the values are the corresponding dictionary of arguments.
        """
        keys = ["dataset", "model", "trainer"]
        params = {}
        for key in keys:
            params[key] = fixed_params[key]
            params[key].update(auto_params[key])

        return {key: params[key] for key in keys}

    @staticmethod
    def _split_config(params):
        """
        Reverses the operation of `get_auto_params` (from a list of parameters (ax-friendly format) to a dictionary of
        dictionaries where keys are dataset, model, and trainer and the values are a dictionary of the corresponding arguments)

        Args:
            params (dict): list of dictionaries where each dictionary specifies a single parameter to be sampled.

        Returns:
            dict: A dictionary of dictionaries where keys are dataset, model, and trainer and the values are the corresponding
            dictionary of to-be-sampled arguments.
        """
        config = dict(dataset={}, model={}, trainer={}, others={})
        for k, v in params.items():
            config[k.split(".")[0]][k.split(".")[1]] = v

        return config

    def train_evaluate(self, auto_params):
        """
        For a given set of parameters, add an entry to the corresponding tables, and populated the trained model
        table for that specific entry.

        Args:
            auto_params (dict): list of dictionaries where each dictionary specifies a single parameter to be sampled.

        """
        config = self._combine_params(self._split_config(auto_params), self.fixed_params)

        # insert the stuff into their corresponding tables
        dataset_hash = make_hash(config["dataset"])
        entry_exists = {
            "dataset_fn": "{}".format(self.fns["dataset"])
        } in self.trained_model_table.dataset_table() and {
            "dataset_hash": "{}".format(dataset_hash)
        } in self.trained_model_table.dataset_table()
        if not entry_exists:
            self.trained_model_table.dataset_table().add_entry(
                self.fns["dataset"],
                config["dataset"],
                dataset_fabrikant=self.architect,
                dataset_comment=self.comment,
            )

        model_hash = make_hash(config["model"])
        entry_exists = {"model_fn": "{}".format(self.fns["model"])} in self.trained_model_table.model_table() and {
            "model_hash": "{}".format(model_hash)
        } in self.trained_model_table.model_table()
        if not entry_exists:
            self.trained_model_table.model_table().add_entry(
                self.fns["model"],
                config["model"],
                model_fabrikant=self.architect,
                model_comment=self.comment,
            )

        trainer_hash = make_hash(config["trainer"])
        entry_exists = {
            "trainer_fn": "{}".format(self.fns["trainer"])
        } in self.trained_model_table.trainer_table() and {
            "trainer_hash": "{}".format(trainer_hash)
        } in self.trained_model_table.trainer_table()
        if not entry_exists:
            self.trained_model_table.trainer_table().add_entry(
                self.fns["trainer"],
                config["trainer"],
                trainer_fabrikant=self.architect,
                trainer_comment=self.comment,
            )

        # get the primary key values for all those entries
        restriction = (
            'dataset_fn in ("{}")'.format(self.fns["dataset"]),
            'dataset_hash in ("{}")'.format(dataset_hash),
            'model_fn in ("{}")'.format(self.fns["model"]),
            'model_hash in ("{}")'.format(model_hash),
            'trainer_fn in ("{}")'.format(self.fns["trainer"]),
            'trainer_hash in ("{}")'.format(trainer_hash),
        )

        # populate the table for those primary keys
        self.trained_model_table().populate(*restriction)

    def gen_params_value(self):
        """
        Generates new values (samples randomly) for each parameter.

        Returns:
            dict: A dictionary containing the parameters whose values should be sampled.
        """
        np.random.seed(None)
        auto_params_val = {}
        for param in self.auto_params:
            if param["type"] == "fixed":
                auto_params_val.update({param["name"]: param["value"]})
            elif param["type"] == "choice":
                auto_params_val.update({param["name"]: np.random.choice(param["values"])})
            elif param["type"] == "range":
                auto_params_val.update({param["name"]: np.random.uniform(*param["bounds"])})

        return auto_params_val

    def run(self):
        """
        Runs the random hyperparameter search, for as many trials as specified.
        """
        n_trials = len(self.trained_model_table.seed_table()) * self.total_trials
        init_len = len(self.trained_model_table())
        while len(self.trained_model_table()) - init_len < n_trials:
            self.train_evaluate(self.gen_params_value())



import copy
import os
import tempfile

import datajoint as dj
import torch

from nnfabrik.templates.trained_model import TrainedModelBase
from nnfabrik.utility.dj_helpers import make_hash, clone_conn, CustomSchema


def my_checkpoint(nnfabrik):
    conn_clone = clone_conn(dj.conn())
    schema_clone = CustomSchema(nnfabrik.schema.database, connection=conn_clone)

    @schema_clone
    class Checkpoint(dj.Manual):
        storage = "minio"

        @property
        def definition(self):
            definition = f"""
            # Checkpoint table
            -> nnfabrik.Trainer
            -> nnfabrik.Dataset
            -> nnfabrik.Model
            -> nnfabrik.Seed
            epoch:                             int          # epoch of creation
            ---
            score:                             float        # current score at epoch
            state:                             attach@{self.storage}  # current state
            ->[nullable] nnfabrik.Fabrikant
            trainedmodel_ts=CURRENT_TIMESTAMP: timestamp    # UTZ timestamp at time of insertion
            """
            return definition

    return Checkpoint


class TrainedModelChkptBase(TrainedModelBase):
    checkpoint_table = None  # assign the output of `my_checkpoint` here (i.e. `Checkpoint`)
    keys = [
        "model_fn",
        "model_hash",
        "dataset_fn",
        "dataset_hash",
        "trainer_fn",
        "trainer_hash",
    ]

    def call_back(self, uid=None, epoch=None, model=None, state=None):
        """
        This method is periodically called by the trainer and is used to save the training state in a remote table.
        Args:
            uid - Unique identifier for the trained model entry
            epoch - the iteration count
            model - current model under training
            state - Additional information provided by the trainer
                action: str = "save"/"last"/"best"
                score: float = 0.0,
                maximize_score: bool = True,
                save_every_n: int = 1,
                keep_last_n: int = 1,
                keep_best_n: int = 1,
                keep_selection: Tuple = (),
        """
        maximize_score = state.pop("maximize_score", True)
        action = state.pop("action", "save")
        if action == "save":
            self.save_epoch(epoch, maximize_score, model, state, uid)
        else:  # restore existing epoch (selected based on action=last/best)
            self.restore_epoch(action, uid, maximize_score, state)

    def restore_epoch(self, action, uid, maximize_score, state):
        with tempfile.TemporaryDirectory() as temp_dir:
            checkpoints = (self.checkpoint_table & uid).fetch(
                "score", "epoch", "state", as_dict=True, download_path=temp_dir
            )
            if not checkpoints:
                return
            if action == "last":  # select last epoch
                last_checkpoints = sorted(checkpoints, key=lambda chkpt: chkpt["epoch"], reverse=False)
                checkpoint = last_checkpoints[-1]
            else:  # select best epoch
                best_checkpoints = sorted(
                    checkpoints,
                    key=lambda chkpt: chkpt["score"],
                    reverse=maximize_score,
                )
                checkpoint = best_checkpoints[0]
            # restore the training state
            state["epoch"] = checkpoint["epoch"]
            state["score"] = checkpoint["score"]
            loaded_state = torch.load(checkpoint["state"])
        for key, state_entry in loaded_state.items():
            if key in state and hasattr(state[key], "load_state_dict"):
                state[key].load_state_dict(state_entry)
            else:
                state[key] = state_entry

    def save_epoch(self, epoch, maximize_score, model, state, uid):
        if epoch % state.pop("save_every_n", 1) != 0:
            return
        score = state.pop("score", 0.0)
        keep_best_n = state.pop("keep_best_n", 1)
        keep_last_n = state.pop("keep_last_n", 1)
        keep_selection = state.pop("keep_selection", ())
        self.add_to_table(epoch, model, score, state, uid)
        self.filter_table(keep_best_n, keep_last_n, keep_selection, maximize_score, uid)

    def filter_table(self, keep_best_n, keep_last_n, keep_selection, maximize_score, uid):
        # fetch all fitting entries from checkpoint table
        checkpoints = (self.checkpoint_table & uid).fetch(
            *self.keys,
            "seed",
            "score",
            "epoch",
            as_dict=True,
        )
        # select checkpoints to be kept
        keep_checkpoints = []
        best_checkpoints = sorted(checkpoints, key=lambda chkpt: chkpt["score"], reverse=maximize_score)
        for c in checkpoints:
            del c["score"]  # restricting with a float is not a good idea -> remove
        keep_checkpoints += best_checkpoints[:keep_best_n]  # w.r.t. performance
        last_checkpoints = sorted(checkpoints, key=lambda chkpt: chkpt["epoch"], reverse=True)
        keep_checkpoints += last_checkpoints[:keep_last_n]  # w.r.t. temporal order
        for chkpt in checkpoints:
            if chkpt["epoch"] in keep_selection:
                keep_checkpoints.append(chkpt)  # keep selected epochs
        # delete the others
        safe_mode = dj.config["safemode"]
        dj.config["safemode"] = False
        ((self.checkpoint_table & uid) - keep_checkpoints).delete(verbose=False)
        dj.config["safemode"] = safe_mode

    def add_to_table(self, epoch, model, score, state, uid):
        with tempfile.TemporaryDirectory() as temp_dir:
            key = copy.deepcopy(uid)
            for k in self.keys:
                if k not in key:
                    key[k] = ""
            key["epoch"] = epoch
            key["score"] = score
            filename = make_hash(uid) + ".pth.tar"
            filepath = os.path.join(temp_dir, filename)
            state["net"] = model.state_dict()
            torch.save(state, filepath)
            key["state"] = filepath
            self.checkpoint_table.insert1(key)  # this is NOT in transaction and thus immediately completes!

    def make(self, key):
        orig_key = copy.deepcopy(key)
        super().make(key)
        # Clean up checkpoints after training:
        trainer_config = (self.trainer_table & orig_key).fetch1("trainer_config")
        if not trainer_config.get("keep_checkpoints"):
            safe_mode = dj.config["safemode"]
            dj.config["safemode"] = False
            (self.checkpoint_table & orig_key).delete(verbose=False)
            print("Deleting intermediate checkpoints...")
            dj.config["safemode"] = safe_mode



import datajoint as dj
import numpy as np
from nnfabrik.main import Model, Dataset, Trainer, Seed, Fabrikant
from .trained_model import TrainedModelBase


class ScoringBase(dj.Computed):
    """
    Inherit from this class and decorate with your own schema to create a functional
    Score table. This serves as a template for all scores that can be computed with a
    TrainedModel and a dataloader.
    Each master table has an attribute that stores the grand average score across all Units.
    The `UnitScore` part table stores the score for all units.
    In order to instantiate a functional table, the default table attriubutes need to be changed.

    This template table is implementing the following logic:
    1) Loading the a trained model from the TrainedModel table. This table needs to have the 'load_model' method.
    2) getting a dataloader. Minimally, the dataloader returns batches of inputs and targets.
        The dataloader will be built by the Datast table of nnfabrik as default.
        This table needs to have the 'get_dataloader' method
    3) passing the model and the dataloader to a scoring function, defined in the class attribute. The function is expected to
        return either the grand average score or the score per unit.
    4) The average scores and unit scores will be stored in the maser/part tables of this template.

    Attributes:
        trainedmodel_table (Datajoint Table) - an instantiation of the TrainedModelBase
        dataset_table (Datajoint Table) - A Dataset table of nnfabrik. By default the Dataset table of the trainedmodel_table.
        measure_function (function object) - a function that given model and dataloader computes a score. The grand average
            score will be stored in the master table, and the unit scores are stored in the .Units part table.
            Refer to lines 47-49 for an example implementation of a measure function.
        measure_dataset (str) - following nnfabrik convention, this string specifies the key for the 'dataloaders'
            object. The dataloaders object has to contain at least ['train', 'validation', 'test'].
            This string determines, on what data tier the score is computed on. Defaults to the test set.
        measure_attribute (str) - name of the non-primary attribute of the master and part tables for the score.
        function_kwargs (dict) - additonal keyword arguments, provided as a mapping, that will be passed to the
            measure function when computing a score.
        cache (object) - A Store that caches models or datasets, so that they don't need to be recomputed for each
            analysis. Ready to use: an instantiation of the FabrikCache (from ..utility.nnf_helper)
    """

    trainedmodel_table = TrainedModelBase
    dataset_table = trainedmodel_table.dataset_table
    measure_dataset = "test"
    measure_attribute = "score"
    function_kwargs = {}
    model_cache = None
    data_cache = None

    @staticmethod
    def measure_function(dataloaders, model, per_unit=True):
        """
        Args:
            dataloaders (dict): an nnfabrik dataloader object
            model (Object): An instance of a TrainedModel
            per_unit (bool): Makes the function either return a unit score, or the grand average score

        Returns:
            (list, array): a score per unit, from which also the grand average score is calculated
        """
        raise NotImplementedError("Scoring Function has to be implemented")

    # table level comment
    table_comment = "A template table for storing results/scores of a TrainedModel"

    @property
    def definition(self):
        definition = """
                # {table_comment}
                -> self.trainedmodel_table
                ---
                {measure_attribute}:      float     # A template for a computed score of a trained model
                {measure_attribute}_ts=CURRENT_TIMESTAMP: timestamp    # UTZ timestamp at time of insertion
                """.format(
            table_comment=self.table_comment, measure_attribute=self.measure_attribute
        )
        return definition

    class Units(dj.Part):
        @property
        def definition(self):
            definition = """
                # Scores for Individual Neurons
                -> master
                unit_index:                   int
                ---
                unit_{measure_attribute}:     float   # A template for a computed unit score        
                """.format(
                measure_attribute=self._master.measure_attribute
            )
            return definition

    def get_model(self, key=None):
        if key is None:
            key = self.fetch1("KEY")

        if self.model_cache is None:
            model = self.trainedmodel_table().load_model(key=key, include_state_dict=True, include_dataloader=False)
        else:
            model = self.model_cache.load(key=key, include_state_dict=True, include_dataloader=False)
        return model

    def get_dataloaders(self, key=None):
        if key is None:
            key = self.fetch1("KEY")
        dataloaders = (
            self.dataset_table().get_dataloader(key=key) if self.data_cache is None else self.data_cache.load(key=key)
        )
        return dataloaders[self.measure_dataset]

    def get_overall_score(self, unit_scores):
        return np.mean(unit_scores)

    def insert_unit_scores(self, key, unit_scores):
        key = key.copy()
        for unit_index, unit_score in enumerate(unit_scores):
            key["unit_index"] = unit_index
            key["unit_{}".format(self.measure_attribute)] = unit_score
            self.Units.insert1(key, ignore_extra_fields=True)

    def make(self, key):
        dataloaders = self.get_dataloaders(key=key)
        model = self.get_model(key=key)
        unit_scores = self.measure_function(model=model, dataloaders=dataloaders, per_unit=True, **self.function_kwargs)

        key[self.measure_attribute] = self.get_overall_score(unit_scores)
        self.insert1(key, ignore_extra_fields=True)
        self.insert_unit_scores(key=key, unit_scores=unit_scores)


class SummaryScoringBase(ScoringBase):
    """
    A template scoring table with the same logic as ScoringBase, but for scores that do not have unit scores, but
    an overall score per model only.
    """

    Units = None

    def make(self, key):
        dataloaders = self.get_dataloaders(key=key)
        model = self.get_model(key=key)
        key[self.measure_attribute] = self.measure_function(
            model=model, dataloaders=dataloaders, **self.function_kwargs
        )
        self.insert1(key, ignore_extra_fields=True)


class MeasuresBase(ScoringBase):
    trainedmodel_table = None
    dataset_table = Dataset

    # table level comment
    table_comment = "A template table for storing measures / descriptive statistics of the Dataset"

    @property
    def definition(self):
        definition = """
                    # {table_comment}
                    -> self.dataset_table
                    ---
                    {measure_attribute}:      float     # A template for a computed score of a trained model
                    {measure_attribute}_ts=CURRENT_TIMESTAMP: timestamp    # UTZ timestamp at time of insertion
                    """.format(
            table_comment=self.table_comment, measure_attribute=self.measure_attribute
        )
        return definition

    class Units(dj.Part):
        @property
        def definition(self):
            definition = """
                # Scores for Individual Neurons
                -> master
                unit_index:                   int     # unit index as extracted by the model
                ---
                unit_{measure_attribute}:     float   # A template for a computed unit score        
                """.format(
                measure_attribute=self._master.measure_attribute
            )
            return definition

    def make(self, key):

        dataloaders = self.get_dataloaders(key=key)
        unit_scores = self.measure_function(dataloaders=dataloaders, per_unit=True, **self.function_kwargs)

        key[self.measure_attribute] = self.get_overall_score(unit_scores)
        self.insert1(key, ignore_extra_fields=True)
        self.insert_unit_scores(key=key, unit_scores=unit_scores)


class SummaryMeasuresBase(MeasuresBase):
    Units = None

    # table level comment
    table_comment = "A template table for storing measures / descriptive statistics of the Dataset"

    def make(self, key):
        dataloaders = self.get_dataloaders(key=key)
        key[self.measure_attribute] = self.measure_function(dataloaders=dataloaders, **self.function_kwargs)
        self.insert1(key, ignore_extra_fields=True)



from . import utility
from functools import partial

from .utility.nnf_helper import split_module_name, dynamic_import
from .utility.nn_helpers import load_state_dict


def resolve_fn(fn_name, default_base):
    """
    Given a string `fn_name`, resolves the name into a callable object. If the name has multiple `.` separated parts, treat all but the last
    as module names to trace down to the final name. If just the name is given, tries to resolve the name in the `default_base` module name context
    with direct eval of `{default_base}.{fn_name}` in this function's context.

    Raises `NameError` if no object matching the name is found and `TypeError` if the resolved object is not callabe.

    When successful, returns the resolved, callable object.
    """
    module_path, class_name = split_module_name(fn_name)

    try:
        fn_obj = (
            dynamic_import(module_path, class_name) if module_path else eval("{}.{}".format(default_base, class_name))
        )
    except NameError:
        raise NameError("Function `{}` does not exist".format(class_name))

    if not callable(fn_obj):
        raise TypeError("The object named {} is not callable.".format(class_name))

    return fn_obj


# provide convenience alias for resolving model, dataset, and trainer
resolve_model = partial(resolve_fn, default_base="models")
resolve_data = partial(resolve_fn, default_base="datasets")
resolve_trainer = partial(resolve_fn, default_base="training")


def get_model(
    model_fn,
    model_config,
    dataloaders=None,
    seed=None,
    state_dict=None,
    strict=True,
    data_info=None,
):
    """
    Resolves `model_fn` and invokes the resolved function with `model_config` keyword arguments as well as the `dataloader` and `seed`.
    Note that the resolved `model_fn` is expected to accept the `dataloader` as the first positional argument and `seed` as a keyword argument.
    If you pass in `state_dict`, the resulting nn.Module instance will be loaded with the state_dict, using appropriate `strict` mode for loading.

    Args:
        model_fn: string name of the model builder function path to be resolved. Alternatively, you can pass in a callable object and no name resolution will be performed.
        model_config: a dictionary containing keyword arguments to be passed into the resolved `model_fn`
        dataloaders: (a dictionary of) dataloaders to be passed into the resolved `model_fn` as the first positional argument
        seed: randomization seed to be passed in to as a keyword argument into the resolved `model_fn`
        state_dict: If provided, the resulting nn.Module object will be loaded with the state_dict before being returned
        strict: Controls the `strict` mode of nn.Module.load_state_dict

    Returns:
        Resulting nn.Module object.
    """

    if isinstance(model_fn, str):
        model_fn = resolve_model(model_fn)

    net = (
        model_fn(dataloaders, seed=seed, **model_config)
        if data_info is None
        else model_fn(dataloaders, data_info=data_info, seed=seed, **model_config)
    )

    if state_dict is not None:
        load_state_dict(net, state_dict)

    return net


def get_data(dataset_fn, dataset_config):
    """
    Resolves `dataset_fn` and invokes the resolved function onto the `dataset_config` configuration dictionary. The resulting
    dataloader will be returned.

    Args:
        dataset_fn: string name of the dataloader function path to be resolved. Alternatively, you can pass in a callable object and no name resolution will be performed.
        dataset_config: a dictionary containing keyword arguments to be passed into the resolved `dataset_fn`

    Returns:
        Result of invoking the resolved `dataset_fn` with `dataset_config` as keyword arguments.
    """
    if isinstance(dataset_fn, str):
        dataset_fn = resolve_data(dataset_fn)

    return dataset_fn(**dataset_config)


def get_trainer(trainer_fn, trainer_config=None):
    """
    If `trainer_fn` string is passed, resolves and returns the corresponding function. If `trainer_config` is passed in,
    a partial function is created with the configuration object expanded.

    Args:
        trainer_fn: string name of the function path to be resolved. Alternatively, you can pass in a callable object and no name resolution will be performed.
        trainer_config: If passed in, a partial function will be created expanding `trainer_config` as the keyword arguments into the resolved trainer_fn

    Returns:
        Resolved trainer function
    """

    if isinstance(trainer_fn, str):
        trainer_fn = resolve_trainer(trainer_fn)

    if trainer_config is not None:
        trainer_fn = partial(trainer_fn, **trainer_config)

    return trainer_fn


def get_all_parts(
    dataset_fn,
    dataset_config,
    model_fn,
    model_config,
    seed=None,
    state_dict=None,
    strict=True,
    trainer_fn=None,
    trainer_config=None,
):

    if seed is not None and "seed" not in dataset_config:
        dataset_config["seed"] = seed  # override the seed if passed in

    dataloaders = get_data(dataset_fn, dataset_config)

    model = get_model(
        model_fn,
        model_config,
        dataloaders,
        seed=seed,
        state_dict=state_dict,
        strict=strict,
    )

    if trainer_fn is not None:
        trainer = get_trainer(trainer_fn, trainer_config)
        return dataloaders, model, trainer
    else:
        return dataloaders, model



from typing import Dict
import numpy as np
import torch
import torch.nn as nn


class MNISTModel(nn.Module):
    def __init__(self, in_dim: int, out_dim: int, h_dim: int = 5) -> None:
        super().__init__()

        self.fc1 = nn.Linear(in_dim, h_dim)
        self.fc2 = nn.Linear(h_dim, out_dim)
        self.nl = nn.ReLU()
        self.softmax = nn.LogSoftmax(dim=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.nl(self.fc1(x))
        return self.softmax(self.fc2(x))


def mnist_model_fn(dataloaders: Dict, seed: int, **config) -> torch.nn.Module:
    """
    Builds a model object for the given config
    Args:
        data_loaders: a dictionary of data loaders
        seed: random seed (e.g. for model initialization)
    Returns:
        Instance of torch.nn.Module
    """
    # get the input and output dimension for the model
    first_input, first_output = next(iter(dataloaders["train"]))
    in_dim = np.prod(first_input.shape[1:])
    out_dim = 10

    torch.manual_seed(seed)  # for reproducibility (almost)
    model = MNISTModel(in_dim, out_dim, h_dim=config.get("h_dim", 5))

    return model
